__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import nseries_fixtures
from aphyt import omron


class TestOnline(nseries_fixtures.BaseNSeries):

    # def __init__(self):
    eip_instance = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.eip_instance = omron.n_series.NSeriesThreadDispatcher()
        cls.eip_instance.connect_explicit('192.168.250.9')
        # cls.eip_instance.register_session()
        # cls.eip_instance.update_variable_dictionary()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.eip_instance.close_explicit()




__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from aphyt.eip import *
from aphyt.cip.cip_datatypes import *
from aphyt.cip.cip import *

def set_bit(data: bytes, bit_position: int):
    """Set an arbitrary bit in a bytes object to 1"""
    length = len(data)
    data_int = int.from_bytes(data, 'little', signed=False)
    mask_int = 2**bit_position
    data_int = data_int | mask_int
    result = data_int.to_bytes(length, 'little', signed=False)
    return result


def clear_bit(data: bytes, bit_position: int):
    """Set an arbitrary bit in a bytes object to 0"""
    length = len(data)
    data_int = int.from_bytes(data, 'little', signed=False)
    mask_int = ~2**bit_position
    data_int = data_int & mask_int
    result = data_int.to_bytes(length, 'little', signed=False)
    return result


def test_bit(data: bytes, bit_position: int):
    """Determine if an arbitrary bit in a bytes object is 1 or 0"""
    data_int = int.from_bytes(data, 'little', signed=False)
    mask_int = 2**bit_position
    if data_int & mask_int:
        result = True
    else:
        result = False
    return result


def clear_bytes(data: bytes):
    """Set bytes object to all 0"""
    length = len(data)
    data = b'\x00' * length
    return data


class F4Series:
    """
    F4Series
    ~~~~~~~

    Class that implements Omron F4 vision sensor specific Ethernet/IP services
    """
    def __init__(self):
        super().__init__()
        self.connected_cip_dispatcher = EIPConnectedCIPDispatcher()
        self.blank_control = 0b0000000000000000
        self.status = b'\x00\x00'
        self.camera_control_register = b'\x00\x00'

    def connect_explicit(self, host):
        self.connected_cip_dispatcher.connect_explicit(host)

    def close_explicit(self):
        self.connected_cip_dispatcher.close_explicit()

    def register_session(self):
        self.connected_cip_dispatcher.register_session()

    def get_control_register(self):
        """
        Method to request the current content of the control register
        :return:
        """
        request_path = address_request_path_segment(class_id=b'\x6d\x00', instance_id=b'\x01', attribute_id=b'\x01')
        reply = self.connected_cip_dispatcher.get_attribute_single_service(request_path)
        return reply.reply_data

    def get_camera_status(self):
        """
        Method to request status register from the camera. This method should be executed prior to using the status
        methods.
        :return:
        """
        request_path = address_request_path_segment(class_id=b'\x6d\x00', instance_id=b'\x01', attribute_id=b'\x02')
        reply = self.connected_cip_dispatcher.get_attribute_single_service(request_path)
        self.status = reply.reply_data
        return self.status

    def status_online(self):
        """Returns true if the status register reports that inspections are running"""
        return test_bit(self.status, 0)

    def status_exposure_busy(self):
        """
        Returns true if the status register reports that the camera is capturing an image. For F440-F it is
        equivalent to acquisition busy
        """
        return test_bit(self.status, 1)

    def status_acquisition_busy(self):
        """Returns true if the status register reports that the camera is acquiring an image"""
        return test_bit(self.status, 2)

    def status_trigger_ready(self):
        """Returns true if the status register reports that the camera is ready to be triggered"""
        return test_bit(self.status, 3)

    def status_error(self):
        """
        Returns true if the status register reports an error has occurred. Set the reset error
        control bit in the command register to reset.
        """
        return test_bit(self.status, 4)

    def status_reset_count_acknowledgement(self):
        """Returns true if the status register reports that the reset count signal was received"""
        return test_bit(self.status, 5)

    def status_execute_command_acknowledgement(self):
        """Returns true if the status register reports that the execute command signal was received"""
        return test_bit(self.status, 7)

    def status_trigger_acknowledgement(self):
        """Returns true if the status register reports the trigger signal was received"""
        return test_bit(self.status, 8)

    def status_inspection_busy(self):
        """Returns true if the status register reports the inspection is busy processing an image"""
        return test_bit(self.status, 9)

    def status_inspection_status(self):
        """
        Returns true if the status register reports the inspection has passed. It is only valid
        when the DataValid bit is high.
        """
        return test_bit(self.status, 10)

    def status_data_valid(self):
        """
        Returns true if the status register reports the inspection 1 is complete. The controller
        should clear this signal by setting the reset data valid once results have been read.
        """
        return test_bit(self.status, 11)

    def get_string(self, number: int):
        """Get the string stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x6c\x00', instance_id=b'\x01', attribute_id=attribute_id)
        reply = self.connected_cip_dispatcher.get_attribute_single_service(request_path)
        reply_string = str(reply.reply_data[4:], 'utf-8')
        return reply_string

    def set_string(self, number: int, value: str):
        """Set the string stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x6c\x00', instance_id=b'\x01', attribute_id=attribute_id)
        data_length = len(value)
        data = data_length.to_bytes(4, 'little', signed=False)
        data = data + bytes(value, 'utf-8')
        self.connected_cip_dispatcher.set_attribute_single_service(request_path, data)

    def get_bool(self, number: int):
        """Get the Boolean stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x68\x00', instance_id=b'\x01', attribute_id=attribute_id)
        reply = self.connected_cip_dispatcher.get_attribute_single_service(request_path)
        if reply.reply_data == b'\x01\x00':
            return True
        else:
            return False

    def set_bool(self, number: int, value: bool):
        """Set the Boolean stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x68\x00', instance_id=b'\x01', attribute_id=attribute_id)
        self.connected_cip_dispatcher.get_attribute_single_service(request_path)
        cip_type = CIPBoolean()
        if value:
            cip_type.from_value(True)
        else:
            cip_type.from_value(False)
        self.connected_cip_dispatcher.set_attribute_single_service(request_path, cip_type.data)

    def get_int(self, number: int):
        """Get the 16-bit Integer stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x69\x00', instance_id=b'\x01', attribute_id=attribute_id)
        reply = self.connected_cip_dispatcher.get_attribute_single_service(request_path)
        value = CIPInteger()
        value.data = reply.reply_data
        return value

    def set_int(self, number: int, value: int):
        """Set the 16-bit Integer stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x69\x00', instance_id=b'\x01', attribute_id=attribute_id)
        cip_type = CIPInteger()
        cip_type.from_value(value)
        self.connected_cip_dispatcher.set_attribute_single_service(request_path, cip_type.data)

    def get_long(self, number: int):
        """Get the 32-bit Integer stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x6a\x00', instance_id=b'\x01', attribute_id=attribute_id)
        reply = self.connected_cip_dispatcher.get_attribute_single_service(request_path)
        value = CIPDoubleInteger()
        value.data = reply.reply_data
        return value

    def set_long(self, number: int, value: int):
        """Set the 32-bit Integer stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x6a\x00', instance_id=b'\x01', attribute_id=attribute_id)
        cip_type = CIPDoubleInteger()
        cip_type.from_value(value)
        self.connected_cip_dispatcher.set_attribute_single_service(request_path, cip_type.data)

    def get_float(self, number: int):
        """Get the Floating Point stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x6b\x00', instance_id=b'\x01', attribute_id=attribute_id)
        reply = self.connected_cip_dispatcher.get_attribute_single_service(request_path)
        value = CIPReal()
        value.data = reply.reply_data
        return value

    def set_float(self, number: int, value: float):
        """Set the Floating Point stored in the camera at the specified attribute number"""
        assert number >= 1
        assert number <= 200
        attribute_id = number.to_bytes(2, 'little', signed=False)
        request_path = address_request_path_segment(
            class_id=b'\x6b\x00', instance_id=b'\x01', attribute_id=attribute_id)
        cip_type = CIPReal()
        cip_type.from_value(value)
        self.connected_cip_dispatcher.set_attribute_single_service(request_path, cip_type.data)

    def send_command_register(self):
        """Send the command register to the camera to execute the control bits that are set"""
        request_path = address_request_path_segment(class_id=b'\x6d\x00', instance_id=b'\x01', attribute_id=b'\x01')
        self.connected_cip_dispatcher.set_attribute_single_service(request_path, self.camera_control_register)
        self.camera_control_register = clear_bytes(self.camera_control_register)
        self.connected_cip_dispatcher.set_attribute_single_service(request_path, self.camera_control_register)

    def go_online(self):
        """Start all inspections running"""
        self.camera_control_register = set_bit(self.camera_control_register, 0)
        self.send_command_register()

    def go_offline(self):
        """Stop all inspections"""
        self.camera_control_register = set_bit(self.camera_control_register, 1)
        self.send_command_register()

    def reset_error(self):
        """Reset error in the status register"""
        self.camera_control_register = set_bit(self.camera_control_register, 4)
        self.send_command_register()

    def reset_count(self):
        """Reset all inspection counts"""
        self.camera_control_register = set_bit(self.camera_control_register, 5)
        self.send_command_register()

    def execute_command(self):
        """Execute command specified by Control.CmdCode"""
        self.camera_control_register = set_bit(self.camera_control_register, 7)
        self.send_command_register()

    def trigger_inspection(self):
        """Trigger inspection 1. The inspection must be configured for  triggered image acquisition"""
        self.get_camera_status()
        while not self.status_trigger_ready():
            self.get_camera_status()
        self.camera_control_register = set_bit(self.camera_control_register, 8)
        self.send_command_register()

    def reset_data_valid(self):
        """
        Reset the Data Valid signal of the status register.
        This should be performed after the data has been read.
        """
        self.camera_control_register = set_bit(self.camera_control_register, 11)
        self.send_command_register()




__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import asyncio

import nseries_fixtures
from aphyt import omron




class AsyncTestOnline(nseries_fixtures.AsyncBaseNSeries):

    # def __init__(self):
    eip_instance = None

    @classmethod
    def setUpClass(cls) -> None:
        # Handling asynchronous set up
        cls.loop = asyncio.new_event_loop()  # create and set new event loop
        asyncio.set_event_loop(cls.loop)
        cls.loop.run_until_complete(cls.asyncSetUpClass())

    @classmethod
    async def asyncSetUpClass(cls):
        # assign measurement object to unit test object
        cls.eip_instance = omron.n_series.AsyncNSeries()
        await cls.eip_instance.connect_explicit('192.168.250.9')
        # await cls.eip_instance.register_session()

    @classmethod
    async def tearDownClass(cls) -> None:
        await cls.eip_instance.close_explicit()
        if cls.loop.is_running():
            cls.loop.stop()
        cls.loop.close()




__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

# from .cip import *
# from .cip_datatypes import *
# from .cip_attributes import *
# from .cip_services import *



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import logging
import os
import struct
import tkinter
from abc import ABC, abstractmethod
from numbers import Number
from tkinter import ttk, TclError
from typing import Dict, Union, SupportsRound
import PIL
from PIL import Image, ImageTk

from aphyt.omron import NSeriesThreadDispatcher, MonitoredVariable


class DispatcherMixin:
    def __init__(self, dispatcher: NSeriesThreadDispatcher = None, **kwargs):
        super().__init__(**kwargs)
        self.dispatcher = dispatcher


class MonitoredVariableWidgetMixin(ABC):
    def __init__(self,
                 dispatcher: NSeriesThreadDispatcher = None,
                 variable_name: str = None,
                 refresh_time=0.5,
                 **kwargs):
        super().__init__(**kwargs)
        self.log = logging.getLogger(__name__)
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.refresh_time = refresh_time
        self.monitored_variable = MonitoredVariable(self.dispatcher, self.variable_name, self.refresh_time)
        self.monitored_variable.bind_to_value(self._value_updated)
        self.dispatcher.add_monitored_variable(self.monitored_variable)

    @abstractmethod
    def _value_updated(self):
        pass


DEFAULT_PADDING = 5


class HMIWidgetFrame(tkinter.Frame):
    def __init__(self, master, background=None, *args, **kwargs):
        super().__init__(master, padx=2, pady=2, *args, **kwargs)
        if background is None:
            try:
                self.config(background=master['background'])
            except TclError:
                pass

    def add_widget_linear(self, widget: tkinter.Widget, vertical=True,
                          pad_x=None, pad_y=None, **kwargs):
        if pad_x is None:
            pad_x = DEFAULT_PADDING
        if pad_y is None:
            pad_y = DEFAULT_PADDING
        if vertical:
            widget.grid(row=self.grid_size()[1] + 1, padx=pad_x,
                        pady=pad_y, sticky='n', **kwargs)
        else:
            widget.grid(column=self.grid_size()[0] + 1, padx=pad_x,
                        pady=pad_y, sticky='w', **kwargs)

    def add_widget_grid(self, widget: tkinter.Widget, row, column,
                        row_span: int = 1, column_span: int = 1,
                        pad_x=None, pad_y=None, **kwargs):
        if pad_x is None:
            pad_x = DEFAULT_PADDING
        if pad_y is None:
            pad_y = DEFAULT_PADDING
        widget.grid(row=row, column=column,
                    rowspan=row_span, columnspan=column_span,
                    padx=pad_x, pady=pad_y, **kwargs)
        if hasattr(widget, 'visibility_variable'):
            if widget.visibility_variable is not None:
                widget.hide()
                widget.visibility_update()


class ImageWidget(tkinter.Label):
    def __init__(self, master=None, image=None,
                 scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.image = None
        self.pressed_image = None
        if not image:
            # image = tkinter.PhotoImage(width=1, height=1)
            path = os.path.dirname(__file__)
            path = os.path.join(path, 'Transparent_Pixel.png')
            self.image = HMIImage(path)
        else:
            self.image = HMIImage(image, scale, scale_x, scale_y)
        super().__init__(master=master, image=self.image.image_tk, borderwidth=0, highlightthickness=0,
                         **kwargs)

    def set_size(self, width: int = 1, height: int = 1):
        self.image.set_size(width, height)
        self.config(image=self.image.image_tk)


class HMIImage:
    def __init__(self, image=None, scale=1.0, scale_x=1.0, scale_y=1.0, **kwargs):
        # ToDo scale, scale_x, scale_y?
        self.image = None
        path = os.path.dirname(__file__)
        if image is None:
            path = os.path.join(path, 'Transparent_Pixel.png')
            self.image = PIL.Image.open(path)
        else:
            # path = os.path.join(path, image)
            self.image = PIL.Image.open(image)
        self.width = int(scale * scale_x * float(self.image.size[0]))
        self.height = int(scale * scale_y * float(self.image.size[1]))
        self.image = self.image.resize((self.width, self.height), PIL.Image.ANTIALIAS)
        self.image_tk = PIL.ImageTk.PhotoImage(self.image)
        # super().__init__(master, image=self.image_tk, compound='center', **kwargs)

    def resize(self, scale=1.0, scale_x=1.0, scale_y=1.0):
        self.width = int(scale * scale_x * float(self.image.size[0]))
        self.height = int(scale * scale_y * float(self.image.size[1]))
        self.image = self.image.resize((self.width, self.height), PIL.Image.ANTIALIAS)
        self.image_tk = PIL.ImageTk.PhotoImage(self.image)

    def set_size(self, width: int = 1, height: int = 1):
        self.width = width
        self.height = height
        self.image = self.image.resize((self.width, self.height), PIL.Image.ANTIALIAS)
        self.image_tk = PIL.ImageTk.PhotoImage(self.image)


class HMIPage(tkinter.Frame):
    def __init__(self, master, background=None, **kwargs):
        super().__init__(master, **kwargs)
        if background is None:
            self.config(background=master['background'])
        self.grid(row=0, column=0, sticky='nsew')
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)


class HMIControllerPageChange(MonitoredVariableWidgetMixin):
    def __init__(self, screen: 'HMIScreen', **kwargs):
        super(HMIControllerPageChange, self).__init__(**kwargs)
        self.screen = screen

    def _value_updated(self):
        self.screen.change_page(self.monitored_variable.value)


class HMIScreen(tkinter.Tk):
    def __init__(self, window_title,
                 dispatcher_host: str,
                 initial_page: str,
                 dictionary_file=None,
                 geometry: str = None,
                 background=None,
                 screen_change_variable=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.initial_page = initial_page
        self.pages = {}
        self.withdraw()
        # path = os.path.dirname(__file__)
        # path = os.path.join(path, 'transparent_icon.ico')
        # self.iconbitmap(path)
        self.title(window_title)
        self.screen_change_variable = screen_change_variable
        if geometry is not None:
            self.geometry(geometry)
        if background is not None:
            self.config(background=background)
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.eip_instance = NSeriesThreadDispatcher(dispatcher_host)
        if dictionary_file is not None:
            self.eip_instance.load_dictionary_file_if_present(dictionary_file)
        else:
            pass
        self.update()
        self.deiconify()

    def show_front_page(self):
        self.change_page(self.initial_page)
        if self.screen_change_variable is not None:
            if type(self.screen_change_variable) == str:
                self.screen_change_variable = HMIControllerPageChange(self,
                                                                      dispatcher=self.eip_instance,
                                                                      variable_name=self.screen_change_variable,
                                                                      refresh_time=0.05)
                self.screen_change_variable.monitored_variable.value = self.initial_page

    def change_page(self, page: str):
        self.pages[page].tkraise()

    def destroy(self):
        self.eip_instance.close_explicit()
        super().destroy()


class MomentaryButtonMixin(DispatcherMixin):
    def __init__(self, variable_name: str = None, **kwargs):
        super().__init__(**kwargs)
        self.variable_name = variable_name

    def _on_press(self, event):
        self.dispatcher.verified_write_variable(self.variable_name, True)

    def _on_release(self, event):
        self.dispatcher.verified_write_variable(self.variable_name, False)


class SetButtonMixin(DispatcherMixin):
    def __init__(self, variable_name: str = None, **kwargs):
        super().__init__(**kwargs)
        self.variable_name = variable_name

    def _on_press(self, event):
        self.dispatcher.verified_write_variable(self.variable_name, True)

    def _on_release(self, event):
        pass


class ResetButtonMixin(DispatcherMixin):
    def __init__(self, variable_name: str = None, **kwargs):
        super().__init__(**kwargs)
        self.variable_name = variable_name

    def _on_press(self, event):
        self.dispatcher.verified_write_variable(self.variable_name, False)

    def _on_release(self, event):
        pass


class ToggleButtonMixin(DispatcherMixin):
    def __init__(self, variable_name: str = None, dispatcher: NSeriesThreadDispatcher = None, **kwargs):
        super().__init__(dispatcher, **kwargs)
        self.variable_name = variable_name

    def _on_press(self, event):
        reply = self.dispatcher.read_variable(self.variable_name)
        if reply:
            self.dispatcher.verified_write_variable(self.variable_name, False)
        else:
            self.dispatcher.verified_write_variable(self.variable_name, True)

    def _on_release(self, event):
        pass


class VariableButtonMixin(DispatcherMixin):
    def __init__(self, variable_name: str = None, value=None, dispatcher: NSeriesThreadDispatcher = None, **kwargs):
        super().__init__(dispatcher, **kwargs)
        self.variable_name = variable_name
        self.value = value

    def _on_press(self, event):
        self.dispatcher.verified_write_variable(self.variable_name, self.value)

    def _on_release(self, event):
        pass


class ImageMultiStateButton(MonitoredVariableWidgetMixin, VariableButtonMixin, tkinter.Button):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str, refresh_time,
                 state_images: Dict[int, str], scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.log = logging.getLogger(__name__)
        self.state_images = state_images
        self.state_images_tk = {}
        for state in self.state_images:
            # image = PIL.Image.open(self.state_images[state])
            # self.state_images_tk[state] = PIL.ImageTk.PhotoImage(image)
            image = HMIImage(state_images[state], scale, scale_x, scale_y)
            self.state_images_tk[state] = image.image_tk
        first_value = list(self.state_images_tk)[0]
        super().__init__(variable_name=variable_name,
                         value=first_value, dispatcher=dispatcher,
                         refresh_time=refresh_time,
                         master=master,
                         **kwargs)
        if background is None:
            self.config(background=master['background'])
        self.value = self.monitored_variable.value
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)
        self._value_updated()

    def _next_value(self):
        temp = list(self.state_images_tk)
        try:
            next_value = temp[temp.index(self.value) + 1]
        except (ValueError, IndexError):
            next_value = temp[0]
        return next_value

    def _value_updated(self):
        try:
            self.config(image=self.state_images_tk[self.monitored_variable.value])
        except KeyError as key_error:
            tkinter.messagebox.showerror('No State Image', 'Is this an available state?')

    def _on_press(self, event):
        self.value = self._next_value()
        super()._on_press(event)


class ImageButtonMixin(tkinter.Button):
    def __init__(self, master=None, dispatcher: NSeriesThreadDispatcher = None,
                 variable_name: str = None,
                 image=None, pressed_image=None,
                 scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.image = None
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.pressed_image = None
        if not image:
            # image = tkinter.PhotoImage(width=1, height=1)
            path = os.path.dirname(__file__)
            path = os.path.join(path, 'Transparent_Pixel.png')
            self.image = HMIImage(path)
        else:
            self.image = HMIImage(image, scale, scale_x, scale_y)
        if pressed_image is None:
            self.pressed_image = self.image
        else:
            self.pressed_image = HMIImage(pressed_image, scale, scale_x, scale_y)
        self.initial_image = self.image
        super().__init__(master=master, image=self.initial_image.image_tk, compound='center', **kwargs)
        if background is None:
            self.config(background=master['background'])
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)

    def _on_press(self, event):
        self.config(image=self.pressed_image.image_tk)

    def _on_release(self, event):
        self.config(image=self.image.image_tk)


class PageSwitchButtonMixin(tkinter.Button):
    def __init__(self, master: tkinter.Widget, page: Union[HMIPage, str], **kwargs):
        super(PageSwitchButtonMixin, self).__init__(master=master, **kwargs)
        self.page = page
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)

    def _on_press(self, event):
        pass

    def _on_release(self, event):
        if type(self.page) == str:
            self.winfo_toplevel().pages[self.page].tkraise()
            if self.winfo_toplevel().screen_change_variable is not None:
                self.winfo_toplevel().screen_change_variable.monitored_variable.value = self.page
        else:
            self.page.tkraise()
            if self.winfo_toplevel().screen_change_variable is not None:
                self.winfo_toplevel().screen_change_variable.monitored_variable.value = self.page.__class__.__name__


class ImagePageSwitchButton(ImageButtonMixin, PageSwitchButtonMixin):
    def __init__(self, master: tkinter.Widget, page: Union[HMIPage, str], scale=1.0,
                 image=None, pressed_image=None, background=None,
                 **kwargs):
        super().__init__(master=master, page=page, image=image, pressed_image=pressed_image, scale=scale, **kwargs)
        if background is None:
            background = master['background']
        self.config(background=background)
        self.config(activebackground=background)
        self.config(highlightthickness=0, borderwidth=0)

    def _on_press(self, event):
        ImageButtonMixin._on_press(self, event)

    def _on_release(self, event):
        ImageButtonMixin._on_release(self, event)
        PageSwitchButtonMixin._on_release(self, event)


class PageSwitchButton(ttk.Button, PageSwitchButtonMixin):
    def __init__(self, master: tkinter.Widget, page: Union[HMIPage, str], **kwargs):
        super().__init__(master, **kwargs)
        self.page = page
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)

    def _on_press(self, event):
        pass

    def _on_release(self, event):
        PageSwitchButtonMixin._on_release(self, event)


class ImageMomentaryButton(MonitoredVariableWidgetMixin, ImageButtonMixin, MomentaryButtonMixin):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str,
                 image=None, pressed_image=None,
                 scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.log = logging.getLogger(__name__)
        super().__init__(master=master,
                         dispatcher=dispatcher,
                         variable_name=variable_name,
                         image=image, pressed_image=pressed_image,
                         scale=scale, scale_x=scale_x, scale_y=scale_y, background=background,
                         **kwargs)
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)

    def _value_updated(self):
        if self.monitored_variable.value:
            self.config(image=self.pressed_image.image_tk)
        else:
            self.config(image=self.image.image_tk)
        
    def _on_press(self, event):
        ImageButtonMixin._on_press(self, event)
        MomentaryButtonMixin._on_press(self, event)

    def _on_release(self, event):
        ImageButtonMixin._on_release(self, event)
        MomentaryButtonMixin._on_release(self, event)


class MomentaryButton(MomentaryButtonMixin, ttk.Button):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str, **kwargs):
        super().__init__(master=master, variable_name=variable_name, dispatcher=dispatcher, **kwargs)
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)


class ImageMomentaryButton(MonitoredVariableWidgetMixin, ImageButtonMixin, MomentaryButtonMixin):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str,
                 image=None, pressed_image=None,
                 scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.log = logging.getLogger(__name__)
        super().__init__(master=master,
                         dispatcher=dispatcher,
                         variable_name=variable_name,
                         image=image, pressed_image=pressed_image,
                         scale=scale, scale_x=scale_x, scale_y=scale_y, background=background,
                         **kwargs)
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)
        self._value_updated()

    def _value_updated(self):
        if self.monitored_variable.value:
            self.config(image=self.pressed_image.image_tk)
        else:
            self.config(image=self.image.image_tk)

    def _on_press(self, event):
        # ImageButtonMixin._on_press(self, event)
        MomentaryButtonMixin._on_press(self, event)

    def _on_release(self, event):
        # ImageButtonMixin._on_release(self, event)
        MomentaryButtonMixin._on_release(self, event)


class ImageSetButton(MonitoredVariableWidgetMixin, ImageButtonMixin, SetButtonMixin):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str,
                 image=None, pressed_image=None,
                 scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.log = logging.getLogger(__name__)
        super().__init__(master=master,
                         dispatcher=dispatcher,
                         variable_name=variable_name,
                         image=image, pressed_image=pressed_image,
                         scale=scale, scale_x=scale_x, scale_y=scale_y, background=background,
                         **kwargs)
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)
        self._value_updated()

    def _value_updated(self):
        if self.monitored_variable.value:
            self.config(image=self.pressed_image.image_tk)
        else:
            self.config(image=self.image.image_tk)

    def _on_press(self, event):
        # ImageButtonMixin._on_press(self, event)
        SetButtonMixin._on_press(self, event)

    def _on_release(self, event):
        SetButtonMixin._on_release(self, event)


class SetButton(SetButtonMixin, tkinter.ttk.Button):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str, **kwargs):
        super().__init__(master=master, variable_name=variable_name, dispatcher=dispatcher, **kwargs)
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)


class ImageResetButton(MonitoredVariableWidgetMixin, ImageButtonMixin, ResetButtonMixin):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str,
                 image=None, pressed_image=None,
                 scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.log = logging.getLogger(__name__)
        super().__init__(master=master,
                         dispatcher=dispatcher,
                         variable_name=variable_name,
                         image=image, pressed_image=pressed_image,
                         scale=scale, scale_x=scale_x, scale_y=scale_y, background=background,
                         **kwargs)
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)
        self._value_updated()

    def _value_updated(self):
        if self.monitored_variable.value:
            self.config(image=self.pressed_image.image_tk)
        else:
            self.config(image=self.image.image_tk)

    def _on_press(self, event):
        # ImageButtonMixin._on_press(self, event)
        ResetButtonMixin._on_press(self, event)

    def _on_release(self, event):
        ResetButtonMixin._on_release(self, event)


class ResetButton(ResetButtonMixin, tkinter.ttk.Button):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str, **kwargs):
        super().__init__(master=master, variable_name=variable_name, dispatcher=dispatcher, **kwargs)
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)


class ImageToggleButton(MonitoredVariableWidgetMixin, ImageButtonMixin, ToggleButtonMixin):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str,
                 image=None, pressed_image=None,
                 scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.log = logging.getLogger(__name__)
        super().__init__(master=master,
                         dispatcher=dispatcher,
                         variable_name=variable_name,
                         image=image, pressed_image=pressed_image,
                         scale=scale, scale_x=scale_x, scale_y=scale_y, background=background,
                         **kwargs)
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)
        self._value_updated()

    def _value_updated(self):
        if self.monitored_variable.value:
            self.config(image=self.pressed_image.image_tk)
        else:
            self.config(image=self.image.image_tk)

    def _on_press(self, event):
        # ImageButtonMixin._on_press(self, event)
        ToggleButtonMixin._on_press(self, event)

    def _on_release(self, event):
        # ImageButtonMixin._on_release(self, event)
        ToggleButtonMixin._on_release(self, event)


class ToggleButton(ToggleButtonMixin, tkinter.ttk.Button):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str, **kwargs):
        super().__init__(master=master, variable_name=variable_name, dispatcher=dispatcher, **kwargs)
        self.bind('<ButtonPress-1>', self._on_press)
        self.bind('<ButtonRelease-1>', self._on_release)


class VisibilityMixin(tkinter.Widget):
    def __init__(self,
                 visibility_variable=None,
                 refresh_time=0.5, **kwargs):
        super().__init__(**kwargs)
        self.visibility_variable = visibility_variable
        self.refresh_time = refresh_time
        self.invisible_widget = ImageWidget()
        self.visible = False
        if self.visibility_variable is not None:
            self.monitored_visibility_variable = MonitoredVariable(self.dispatcher,
                                                                   self.visibility_variable,
                                                                   self.refresh_time)
            self.monitored_visibility_variable.bind_to_value(self.visibility_update)
            # self.after(50, self.wait_visibility)
            self.visibility_update()

    def visibility_update(self):
        if self.monitored_visibility_variable.value:
            self.show()
        else:
            self.hide()

    def show(self):
        self.invisible_widget.grid_remove()
        self.grid()

    def hide(self):
        self.invisible_widget.set_size(self.winfo_reqwidth(),
                                       self.winfo_reqheight())
        self.invisible_widget.grid(self.grid_info())
        self.grid_remove()


class ImageMultiStateLamp(MonitoredVariableWidgetMixin, VisibilityMixin, tkinter.Label):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name, refresh_time,
                 state_images: Dict[int, str], scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.log = logging.getLogger(__name__)
        self.dispatcher = dispatcher
        self.state_images = state_images
        self.state_images_tk = {}
        for state in self.state_images:
            image = HMIImage(self.state_images[state], scale, scale_x, scale_y)
            self.state_images_tk[state] = image.image_tk
        super().__init__(master=master, dispatcher=dispatcher, variable_name=variable_name,
                         refresh_time=refresh_time, **kwargs)
        if background is None:
            self.config(background=master['background'])
        self._value_updated()

    def _value_updated(self):
        try:
            self.config(image=self.state_images_tk[self.monitored_variable.value])
        except KeyError as key_error:
            tkinter.messagebox.showerror('No State Image', 'Is this an available state?')


class ImageLamp(MonitoredVariableWidgetMixin, tkinter.Label):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name, refresh_time,
                 image_on, image_off, scale=1.0, scale_x=1.0, scale_y=1.0, background=None, **kwargs):
        self.log = logging.getLogger(__name__)
        self.image_on = HMIImage(image_on, scale, scale_x, scale_y)
        self.image_off = HMIImage(image_off, scale, scale_x, scale_y)
        super().__init__(master=master, dispatcher=dispatcher,
                         variable_name=variable_name, refresh_time=refresh_time, **kwargs)
        if background is None:
            self.config(background=master['background'])
        self._value_updated()

    def _value_updated(self):
        self.log.debug(f'ImageLamp is updated')
        if self.monitored_variable.value:
            self.config(image=self.image_on.image_tk)
        else:
            self.config(image=self.image_off.image_tk)


class DataDisplay(tkinter.ttk.Label):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str,
                 decimal_places=None, rollover=None, delay: int = 100, **kwargs):
        super().__init__(master, **kwargs)
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.decimal_places = decimal_places
        self.rollover = rollover
        self.delay = delay
        self.monitored_variable = MonitoredVariable(self.dispatcher, self.variable_name)
        self._update_data()
        self.monitored_variable.bind_to_value(self._update_data)

    def _update_data(self):
        data = self.monitored_variable.value
        if isinstance(data, SupportsRound) and self.decimal_places is not None:
            data = round(data, self.decimal_places)
            if self.rollover is not None and data >= self.rollover:
                data = 0.0
        data = str(data)
        self.config(text=data)


class DataEdit(ttk.Entry):
    def __init__(self, master, dispatcher: NSeriesThreadDispatcher, variable_name: str, **kwargs):
        super().__init__(master, **kwargs)
        self.dispatcher = dispatcher
        self.variable_name = variable_name
        self.current_string = None
        self.type = None
        self._get_data()
        self.delay = None
        self.dispatcher.connection_status.bind_to_session_status(self._get_data)
        self.validate_command = (self.register(self.validate_entry), '%d', '%i', '%P', '%s', '%S', '%v', '%V', '%W')
        self.config(validate='focus', validatecommand=self.validate_command)
        self.bind("<Return>", (lambda event: self._enter_function()))

    def _get_data(self):
        if self.dispatcher.connection_status.has_session:
            data = self.dispatcher.read_variable(self.variable_name)
            self.current_string = str(data)
            self.type = type(data)
            self.delete(0, 'end')
            self.insert(0, self.current_string)

    def _enter_function(self):
        # print(f'Enter pressed in {self.variable_name}')
        if self.validate():
            self.current_string = self.get()
        else:
            self.delete(0, 'end')
            self.insert(0, self.current_string)
            tkinter.messagebox.showerror('Write Error', 'Could not write value. Is it the correct type?')

    def validate_entry(self, d, i, P, s, S, v, V, W):
        success = False
        try:
            value = self.type(self.get())
            self.dispatcher.verified_write_variable(self.variable_name, value)
            success = True
        except struct.error as error:
            success = False
        except ValueError as error:
            success = False
        return success



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from aphyt.cip.cip_datatypes import *
from aphyt.eip import *
import unittest
from unittest.mock import Mock
from unittest.mock import patch


class TestCipDataTypes(unittest.TestCase):
    def test_cip_boolean_true(self):
        cip_boolean = CIPBoolean()
        cip_boolean.from_value(True)
        self.assertEqual(cip_boolean.data, b'\x01\x00')

    def test_cip_boolean_false(self):
        cip_boolean = CIPBoolean()
        cip_boolean.from_value(False)
        self.assertEqual(cip_boolean.data, b'\x00\x00')

    def test_cip_boolean_true_bytes(self):
        cip_boolean = CIPBoolean()
        cip_boolean.from_value(True)
        self.assertEqual(cip_boolean.value(), True)

    def test_cip_boolean_false_bytes(self):
        cip_boolean = CIPBoolean()
        cip_boolean.from_value(False)
        self.assertEqual(cip_boolean.value(), False)

    def test_cip_short_integer_low(self):
        cip_short_integer = CIPShortInteger()
        cip_short_integer.from_value(-128)
        self.assertEqual(cip_short_integer.data, b'\x80')

    def test_cip_short_integer_high(self):
        cip_short_integer = CIPShortInteger()
        cip_short_integer.from_value(127)
        self.assertEqual(cip_short_integer.data, b'\x7f')

    def test_cip_integer_low(self):
        cip_short_integer = CIPInteger()
        cip_short_integer.from_value(-32768)
        self.assertEqual(cip_short_integer.data, b'\x00\x80')

    def test_cip_integer_high(self):
        cip_integer = CIPInteger()
        cip_integer.from_value(32767)
        self.assertEqual(cip_integer.data, b'\xff\x7f')

    def test_cip_double_integer_low(self):
        cip_double_integer = CIPDoubleInteger()
        cip_double_integer.from_value(-2147483648)
        self.assertEqual(cip_double_integer.data, b'\x00\x00\x00\x80')

    def test_cip_double_integer_high(self):
        cip_double_integer = CIPDoubleInteger()
        cip_double_integer.from_value(2147483647)
        self.assertEqual(cip_double_integer.data, b'\xff\xff\xff\x7f')

    def test_cip_long_integer_low(self):
        cip_long_integer = CIPLongInteger()
        cip_long_integer.from_value(-9223372036854775808)
        self.assertEqual(cip_long_integer.data, b'\x00\x00\x00\x00\x00\x00\x00\x80')

    def test_cip_long_integer_high(self):
        cip_long_integer = CIPLongInteger()
        cip_long_integer.from_value(9223372036854775807)
        self.assertEqual(cip_long_integer.data, b'\xff\xff\xff\xff\xff\xff\xff\x7f')

    def test_cip_unsigned_short_integer_low(self):
        cip_short_integer = CIPUnsignedShortInteger()
        cip_short_integer.from_value(0)
        self.assertEqual(cip_short_integer.data, b'\x00')

    def test_cip_unsigned_short_integer_high(self):
        cip_short_integer = CIPUnsignedShortInteger()
        cip_short_integer.from_value(255)
        self.assertEqual(cip_short_integer.data, b'\xff')

    def test_cip_unsigned_integer_low(self):
        cip_short_integer = CIPUnsignedInteger()
        cip_short_integer.from_value(0)
        self.assertEqual(cip_short_integer.data, b'\x00\x00')

    def test_cip_unsigned_integer_high(self):
        cip_integer = CIPUnsignedInteger()
        cip_integer.from_value(65535)
        self.assertEqual(cip_integer.data, b'\xff\xff')

    def test_cip_unsigned_double_integer_low(self):
        cip_double_integer = CIPUnsignedDoubleInteger()
        cip_double_integer.from_value(0)
        self.assertEqual(cip_double_integer.data, b'\x00\x00\x00\x00')

    def test_cip_unsigned_double_integer_high(self):
        cip_double_integer = CIPUnsignedDoubleInteger()
        cip_double_integer.from_value(4294967295)
        self.assertEqual(cip_double_integer.data, b'\xff\xff\xff\xff')

    def test_cip_unsigned_long_integer_low(self):
        cip_long_integer = CIPUnsignedLongInteger()
        cip_long_integer.from_value(0)
        self.assertEqual(cip_long_integer.data, b'\x00\x00\x00\x00\x00\x00\x00\x00')

    def test_cip_unsigned_long_integer_high(self):
        cip_long_integer = CIPUnsignedLongInteger()
        cip_long_integer.from_value(18446744073709551615)
        self.assertEqual(cip_long_integer.data, b'\xff\xff\xff\xff\xff\xff\xff\xff')

    def test_cip_real_zero(self):
        cip_real = CIPReal()
        cip_real.data = b'\x00\x00\x00\x00'
        self.assertAlmostEqual(cip_real.value(), 0.0)

    def test_cip_real_low(self):
        cip_real = CIPReal()
        cip_real.data = b'\xff\xff\x7f\xff'
        self.assertAlmostEqual(cip_real.value(), -3.4028234663852886e+38)

    def test_cip_real_high(self):
        cip_real = CIPReal()
        cip_real.data = b'\xff\xff\x7f\x7f'
        self.assertAlmostEqual(cip_real.value(), 3.4028234663852886e+38)

    def test_eip_commands(self):
        eip_test = EIPConnectedCIPDispatcher()
        eip_test.send_command = Mock()
        eip_test.list_services('')
        args = eip_test.send_command.call_args
        self.assertEqual(args.args[0].bytes(),
                         b'\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' +
                         b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')

    def test_structure(self):
        cip_structure = CIPStructure()
        cip_structure.variable_name = "Yes"
        cip_structure.variable_type_name = 'LREALisREAL'
        cip_structure.add_member('lreal_member', CIPLongReal())
        cip_structure['lreal_member'] = 34.12121212
        self.assertEqual(cip_structure.data, b'\xb1\xa3\xf5\xe0\x83\x0fA@')

    def test_structure2(self):
        cip_structure = CIPStructure()
        cip_structure.variable_name = "Yes"
        cip_structure.variable_type_name = 'LREALisREAL'
        cip_structure.add_member('lreal_member', CIPLongReal())
        cip_structure['lreal_member'].data = b'\xb1\xa3\xf5\xe0\x83\x0fA@'
        self.assertAlmostEqual(cip_structure['lreal_member'].value(), 34.12121212)

    def test_structure3(self):
        """
        Given a CIPStructure, a variable is updated, the structure data must be updated accordingly.
        """
        cip_structure = CIPStructure()
        cip_structure.variable_name = "Yes"
        cip_structure.variable_type_name = 'YesType'
        cip_structure.add_member('bool_1', CIPBoolean())
        cip_structure['bool_1'] = True
        cip_structure.add_member('bool_2', CIPBoolean())
        cip_structure['bool_2'] = False
        cip_structure.add_member('bool_3', CIPBoolean())
        cip_structure['bool_3'] = True

        self.assertEqual(cip_structure.data, b'\x01\x00\x00\x00\x01\x00')  # This is just to be sure

        cip_structure['bool_3'] = False

        self.assertEqual(cip_structure.data, b'\x01\x00\x00\x00\x00\x00')  # This is the real test assert

    def test_structure_contents_update_updates_data(self):
        """

        Given a CIPStructure, a variable is updated, the structure data must be updated accordingly.

        """
        cip_structure = CIPStructure()
        cip_structure.variable_name = "Yes"
        cip_structure.variable_type_name = 'LREALisREAL'
        cip_structure.add_member('lreal_member', CIPLongReal())
        cip_structure['lreal_member'] = 0.00123

        self.assertEqual(cip_structure.data, b'\xd7\x86\x8a\x71\xfe\x26\x54\x3f')  # This is just to be sure

        cip_structure['lreal_member'] = 34.12121212

        self.assertEqual(cip_structure.data, b'\xb1\xa3\xf5\xe0\x83\x0fA@')  # This is the real test assert

    def test_structure_with_structure_inside(self):
        """

        Given a CIPStructure that has a CIPStructure inside, the structure data must exist.

        """
        cip_structure = CIPStructure()
        cip_structure.size = 6
        cip_structure.variable_name = "outer_struct"
        cip_structure.variable_type_name = 'OuterStruct'

        inner_struct_key = 'inner_struct'
        cip_structure.add_member(inner_struct_key, CIPStructure())
        cip_structure[inner_struct_key].size = 6
        cip_structure[inner_struct_key].variable_type_name = 'StructWithBoolsInside'
        cip_structure[inner_struct_key].add_member('bool_1', CIPBoolean())
        cip_structure[inner_struct_key]['bool_1'] = False
        cip_structure[inner_struct_key].add_member('bool_2', CIPBoolean())
        cip_structure[inner_struct_key]['bool_2'] = True
        cip_structure[inner_struct_key].add_member('bool_3', CIPBoolean())
        cip_structure[inner_struct_key]['bool_3'] = False

        cip_structure.from_value(cip_structure)
        print(f"{cip_structure=}")
        # output -> cip_structure=outer_struct { type: OuterStruct | members: {'inner_struct':
        # { type: StructWithBoolsInside | members: {'bool_1': False, 'bool_2': True, 'bool_3': False} }} }

        print(f'before {binascii.hexlify(cip_structure.data, "-")=}')
        # output -> before binascii.hexlify(cip_structure.data, "-")=b''

        cip_structure.from_value(cip_structure)  # are we supposed to do this to update cip_structure.data ?

        print(f'after {binascii.hexlify(cip_structure.data, "-")=}')
        # output -> after binascii.hexlify(cip_structure.data, "-")=b'00-00-01-00-00-00'

        self.assertEqual(cip_structure.data, b'\x00\x00\x01\x00\x00\x00')  # This is just to be sure

        cip_structure[inner_struct_key]['bool_3'] = True

        # Using the following line does not work, neither using it (but gives different result).
        cip_structure.from_value(cip_structure)  # are we supposed to do this to update cip_structure.data ?

        self.assertEqual(cip_structure.data, b'\x00\x00\x01\x00\x01\x00')  # This is the real test assert



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import asyncio
import binascii
import concurrent.futures
import pickle
import socket
import struct
import sys
import threading
from signal import signal, SIGINT
from abc import ABC, abstractmethod
from aphyt.eip import *
from aphyt.cip.cip_datatypes import *
from aphyt.cip.cip import *
import logging
from threading import Thread


class VariableTypeObjectReply(CIPReply):
    """
    CIP Reply from the Get Attribute All service to Variable Type Object Class Code 0x6C adding descriptive properties
    Omron Vendor specific
    """

    def __init__(self, reply_bytes: bytes):
        super().__init__(reply_bytes=reply_bytes)

    @property
    def size_in_memory(self):
        return struct.unpack("<L", self.reply_data[0:4])[0]

    @property
    def size(self):
        return self.size_in_memory

    @property
    def cip_data_type(self):
        return self.reply_data[5:6]

    @property
    def cip_data_type_of_array(self):
        return self.reply_data[6:7]

    @property
    def array_dimension(self):
        return struct.unpack("<B", self.reply_data[7:8])[0]

    @property
    def number_of_elements(self):
        """Number of elements in each dimension of  the array"""
        dimension_size_list = []
        for i in range(self.array_dimension):
            dimension_size = struct.unpack("<L", self.reply_data[8 + i * 4:8 + i * 4 + 4])[0]
            dimension_size_list.append(dimension_size)
        return dimension_size_list

    @property
    def number_of_members(self):
        return struct.unpack("<H", self.reply_data[8 + self.array_dimension * 4:10 + self.array_dimension * 4])[0]

    @property
    def crc_code(self):
        return struct.unpack("<H", self.reply_data[14 + self.array_dimension * 4:16 + self.array_dimension * 4])[0]

    @property
    def variable_type_name_length(self):
        return struct.unpack("<B", self.reply_data[16 + self.array_dimension * 4:17 + self.array_dimension * 4])[0]

    @property
    def padding(self):
        if self.variable_type_name_length % 2 == 0:
            return 1
        else:
            return 0

    @property
    def variable_type_name(self):
        return \
            self.reply_data[17 + self.array_dimension * 4:
                            17 + self.array_dimension * 4 + self.variable_type_name_length]

    @property
    def next_instance_id(self):
        return \
            self.reply_data[self.padding + 17 + self.array_dimension * 4 + self.variable_type_name_length:
                            self.padding + 17 + self.array_dimension * 4 + self.variable_type_name_length + 4]

    @property
    def nesting_variable_type_instance_id(self):
        return \
            self.reply_data[self.padding + 21 + self.array_dimension * 4 + self.variable_type_name_length:
                            self.padding + 21 + self.array_dimension * 4 + self.variable_type_name_length + 4]

    @property
    def start_array_elements(self):
        """Number of elements in each dimension of  the array"""
        array_start_list = []
        for i in range(self.array_dimension):
            array_start = \
                struct.unpack("<L",
                              self.reply_data[self.padding + 25 +
                                              self.array_dimension * 4 +
                                              self.variable_type_name_length:
                                              self.padding + 25 +
                                              self.array_dimension * 4 +
                                              self.variable_type_name_length + 4])[0]
            array_start_list.append(array_start)
        return array_start_list


class VariableObjectReply(CIPReply):
    """
    CIP Reply from the Get Attribute All service to Variable Object Class Code 0x6B adding descriptive properties
    Omron Vendor specific
    """

    def __init__(self, reply_bytes: bytes):
        super().__init__(reply_bytes=reply_bytes)

    @property
    def size(self):
        return struct.unpack("<L", self.reply_data[0:4])[0]

    @property
    def cip_data_type(self):
        return self.reply_data[4:5]

    @property
    def cip_data_type_of_array(self):
        return self.reply_data[5:6]

    @property
    def array_dimension(self):
        # One byte of padding after this. Skip to byte 8
        return struct.unpack("<B", self.reply_data[6:7])[0]

    @property
    def number_of_elements(self):
        """Number of elements in each dimension of  the array"""
        dimension_size_list = []
        for i in range(self.array_dimension):
            dimension_size = struct.unpack("<L", self.reply_data[8 + i * 4:8 + i * 4 + 4])[0]
            dimension_size_list.append(dimension_size)
        return dimension_size_list

    @property
    def bit_number(self):
        return struct.unpack("<B", self.reply_data[16 + self.array_dimension * 4:17 + self.array_dimension * 4])[0]

    @property
    def variable_type_instance_id(self):
        return self.reply_data[20 + self.array_dimension * 4:24 + self.array_dimension * 4]

    @property
    def start_array_elements(self):
        """Number of elements in each dimension of  the array"""
        array_start_list = []
        for i in range(self.array_dimension):
            array_start = \
                struct.unpack("<L",
                              self.reply_data[24 + self.array_dimension * 4:
                                              24 + self.array_dimension * 4 + 4])[0]
            array_start_list.append(array_start)
        return array_start_list


class InstanceIDAttributes:
    def __init__(self, data: bytes):
        self.data = data

    @property
    def data_length(self):
        return struct.unpack('<H', self.data[0:2])[0]

    @property
    def class_id(self):
        return self.data[2:4]

    @property
    def instance_id(self):
        return self.data[4:8]

    def tag_name_length(self):
        return struct.unpack('<B', self.data[8:9])[0]

    def tag_name(self):
        return str(self.data[9: 9 + self.tag_name_length()], 'utf-8')


class VariableNameAttributeAllReply(CIPReply):
    """
    CIP Reply from the Get Attribute All service to Tag Name adding descriptive properties
    Omron Vendor specific
    """

    def __init__(self, reply_bytes: bytes):
        super().__init__(reply_bytes=reply_bytes)

    @property
    def cip_data_type(self):
        return self.reply_data[4:5]

    @property
    def instance_id(self):
        """
        Instance ID will is where the
        :return:
        """
        return self.reply_data[8:12]

    @property
    def variable_type_id(self):
        return self.reply_data[12:16]


class SimpleDataSegmentRequest:
    """
    A simple data segment request is the Omron specific format for requesting data that will not fit
    in a single CIP message. This is used for strings, arrays and structures
    """

    def __init__(self, offset, size):
        self.simple_data_type_code = b'\x80'
        self.segment_length = b'\x03'  # Fixed and in words
        self.offset = offset
        self.size = size

    def bytes(self):
        return \
                self.simple_data_type_code + self.segment_length + \
                struct.pack("<L", self.offset) + struct.pack("<H", self.size)


class NSeries:
    """
    NSeries
    ~~~~~~~

    Class to read and write data to an Omron N-Series controller. This implementation passes messages to the
    AsyncNSeries class running an event loop so that the asynchronous code can be executed in a synchronous
    program.
    """
    def __init__(self, host=None, timeout=None):
        super().__init__()
        self.derived_data_type_dictionary = {}
        self._instance = AsyncNSeries()
        self.host = host
        self.timeout = timeout
        update_data_type_dictionary(self._instance.connected_cip_dispatcher.data_type_dictionary)
        if self.host is not None:
            self.connect_explicit(self.host, self.timeout)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_explicit()
        self._instance.stop_loop()

    @property
    def connected_cip_dispatcher(self):
        return self._instance.connected_cip_dispatcher

    def connect_explicit(self, host, connection_timeout: float = None):
        self._instance.create_sync_entry()
        future = asyncio.run_coroutine_threadsafe(
            self._instance.connect_explicit(host, connection_timeout), self._instance.loop)
        future.result()

    def close_explicit(self):
        future = asyncio.run_coroutine_threadsafe(
            self._instance.close_explicit(), self._instance.loop)
        future.result()

    def register_session(self):
        # No longer required, but put in for backwards compatibility
        pass

    def update_variable_dictionary(self):
        future = asyncio.run_coroutine_threadsafe(
            self._instance.update_variable_dictionary(), self._instance.loop)
        future.result()

    def variable_list(self):
        future = asyncio.run_coroutine_threadsafe(
            self._instance.variable_list(), self._instance.loop)
        return future.result()

    def user_variable_list(self):
        future = asyncio.run_coroutine_threadsafe(
            self._instance.user_variable_list(), self._instance.loop)
        return future.result()

    def system_variable_list(self):
        future = asyncio.run_coroutine_threadsafe(
            self._instance.system_variable_list(), self._instance.loop)
        return future.result()

    def read_variable(self, variable_name: str):
        future = asyncio.run_coroutine_threadsafe(
            self._instance.read_variable(variable_name), self._instance.loop)
        return future.result()

    def write_variable(self, variable_name: str, data):
        future = asyncio.run_coroutine_threadsafe(
            self._instance.write_variable(variable_name, data), self._instance.loop)
        return future.result()

    def verified_write_variable(self, variable_name: str, data, retry_count: int = 2):
        future = asyncio.run_coroutine_threadsafe(
            self._instance.verified_write_variable(variable_name, data, retry_count), self._instance.loop)
        return future.result()


class AsyncNSeries:
    """
    AsyncNSeries
    ~~~~~~~

    Asynchronous class that implements Omron N-Series specific Ethernet/IP services in addition to Ethernet/IP services
    and CIP services common to most Ethernet/IP devices
    """
    MAXIMUM_LENGTH = 502  # UCMM maximum length is 502 bytes

    def __init__(self, host=None, timeout=None):
        super().__init__()
        self.derived_data_type_dictionary = {}
        self.connected_cip_dispatcher = AsyncEIPConnectedCIPDispatcher()
        self.host = host
        self.timeout = timeout
        update_data_type_dictionary(self.connected_cip_dispatcher.data_type_dictionary)
        self.loop = None
        self.thread = None
        self.done = False

    def start_loop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def create_sync_entry(self):
        self.loop = asyncio.new_event_loop()
        try:
            self.thread = Thread(target=self.start_loop)
            self.thread.start()

        finally:
            pass

    def stop_loop(self):
        self.loop.call_soon_threadsafe(self.loop.stop)
        # self.thread.join()

    def run_method(self, method, positional_arguments, keyword_arguments):
        future = asyncio.run_coroutine_threadsafe(method(*positional_arguments, **keyword_arguments), self.loop)
        return future.result()

    async def __aenter__(self):
        if self.host is not None:
            await self.connect_explicit(self.host)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close_explicit()

    async def connect_explicit(self, host, connection_timeout: float = None):
        await self.connected_cip_dispatcher.connect_explicit(host, connection_timeout)
        await self.register_session()

    async def close_explicit(self):
        await self.connected_cip_dispatcher.close_explicit()
        if self.loop is not None:
            self.stop_loop()

    async def register_session(self):
        await self.connected_cip_dispatcher.register_session()

    async def get_instance_list_service(self, tag_request_path: bytes, data: bytes):
        """
        Omron specific CIP service
        :param tag_request_path:
        :param data:
        :return:
        """
        get_instance_list_request = CIPRequest(b'\x5f', tag_request_path, data)
        return await self.connected_cip_dispatcher.execute_cip_command(get_instance_list_request)

    async def update_derived_data_type_dictionary(self, display=False):
        number_of_entries = await self._get_number_of_derived_data_types()
        for index in range(1, number_of_entries + 1):
            reply = await self._get_variable_type_object(index)
            if display:
                print('index: %s '
                      'size: %s '
                      'type: %s '
                      'array type: %s '
                      'dimensions: %s '
                      'elements: %s '
                      'number of members: %s '
                      'crc: %s '
                      'variable name length: %s '
                      'variable type name: %s '
                      'next instance id: %s '
                      'nesting variable type instance id %s '
                      'start array elements %s' %
                      (str(index),
                       str(reply.size_in_memory),
                       str(reply.cip_data_type),
                       str(reply.cip_data_type_of_array),
                       str(reply.array_dimension),
                       str(reply.number_of_elements),
                       str(reply.number_of_members),
                       str(reply.crc_code),
                       str(reply.variable_type_name_length),
                       str(reply.variable_type_name),
                       str(reply.next_instance_id),
                       str(reply.nesting_variable_type_instance_id),
                       str(reply.start_array_elements)
                       ))

    async def update_variable_dictionary(self):
        """
        Make sure the variable dictionary is populated with the
        latest variable and datatype information from controller
        :return:
        """
        update_data_type_dictionary(self.connected_cip_dispatcher.data_type_dictionary)
        user_instance_list = await self._get_instance_list_subset(True)
        system_instance_list = await self._get_instance_list_subset(False)
        for instance in user_instance_list:
            variable = instance.tag_name()
            variable_cip_datatype = await self._get_instance_from_variable_name(variable)
            variable_cip_datatype.variable_name = str(variable)
            self.connected_cip_dispatcher.variables.update({variable: variable_cip_datatype})
            self.connected_cip_dispatcher.user_variables.update({variable: variable_cip_datatype})

        for instance in system_instance_list:
            variable = instance.tag_name()
            variable_cip_datatype = await self._get_instance_from_variable_name(variable)
            variable_cip_datatype.variable_name = str(variable)
            self.connected_cip_dispatcher.variables.update({variable: variable_cip_datatype})
            self.connected_cip_dispatcher.system_variables.update({variable: variable_cip_datatype})

    async def variable_list(self):
        """
        Return list of variables in the variable dictionary. It will be empty unless the
        update_variable_dictionary, read_variable or write_variable methods have been run.
        :return:
        """
        return list(self.connected_cip_dispatcher.variables)

    async def user_variable_list(self):
        """
        Return list of user variables in the variable dictionary. It will be empty unless the
        update_variable_dictionary, read_variable or write_variable methods have been run.
        :return:
        """
        return list(self.connected_cip_dispatcher.user_variables)

    async def system_variable_list(self):
        """
        Return list of system variables in the variable dictionary. It will be empty unless the
        update_variable_dictionary, read_variable or write_variable methods have been run.
        :return:
        """
        return list(self.connected_cip_dispatcher.system_variables)

    async def variable_dictionary(self, update_contents: bool = False):
        """
        Returns the current variable dictionary. It will only contain variables added by
        update_variable_dictionary, read_variable or write_variable methods.

        The values will only be updated if update_contents parameter is True.
        This is a very expensive operation and should be considered carefully.
        :return:
        """
        if update_contents:
            for variable in await self.variable_list():
                await self.read_variable(variable)
        return self.connected_cip_dispatcher.variables

    async def user_variable_dictionary(self, update_contents: bool = False):
        """
        Returns user variable dictionary. It will only contain variables added by
        update_variable_dictionary, read_variable or write_variable methods.

        The values will only be updated if update_contents parameter is True.
        This is a very expensive operation and should be considered carefully.
        :return:
        """
        if update_contents:
            for variable in await self.user_variable_list():
                await self.read_variable(variable)
        return self.connected_cip_dispatcher.user_variables

    async def system_variable_dictionary(self, update_contents: bool = False):
        """
        Returns system variable dictionary. It will only contain variables added by
        update_variable_dictionary, read_variable or write_variable methods.

        The values will only be updated if update_contents parameter is True.
        This is a very expensive operation and should be considered carefully.
        :return:
        """
        if update_contents:
            for variable in await self.system_variable_list():
               await self.read_variable(variable)
        return self.connected_cip_dispatcher.system_variables

    async def save_current_dictionary(self, file_name: str):
        with (open(file_name, "wb")) as f:
            pickle.dump(self.connected_cip_dispatcher.variables, f)

    async def load_dictionary_file(self, file_name: str):
        with (open(file_name, "rb")) as f:
            self.connected_cip_dispatcher.variables = pickle.load(f)

    async def load_dictionary_file_if_present(self, file_name: str):
        """
        If the supplied file name exists, load it as the controller's dictionary, and if not, pull it
        from the controller and save the dictionary as the supplied file name.

        This allows the programmer to use a file instead of network operation to quickly populate the variable
        dictionary, and if the controller's variable dictionary changes, the user can get the new values
        by simply deleting the file and it will be repopulated on the next run.
        :param file_name:
        :return:
        """
        try:
            with (open(file_name, "rb")) as f:
                self.connected_cip_dispatcher.variables = pickle.load(f)
        except IOError as err:
            await self.update_variable_dictionary()
            await self.save_current_dictionary(file_name)

    async def _structure_instance_from_variable_type_object(
            self, variable_type_object: VariableTypeObjectReply) -> CIPStructure:
        """
        This method recursively builds a CIP Structure from a Variable Type Object that represents a
        structure definition
        :param variable_type_object:
        :return:
        """
        nesting_id = variable_type_object.nesting_variable_type_instance_id
        nesting_id = int.from_bytes(nesting_id, 'little')
        nested_variable_type_object = await self._get_variable_type_object(nesting_id)
        cip_datatype_instance = CIPStructure()
        cip_datatype_instance.instance_id = nesting_id
        if variable_type_object.number_of_members == 0:
            # This is a case where it is referencing a structure elsewhere
            cip_datatype_instance.variable_type_name = str(nested_variable_type_object.variable_type_name, 'utf-8')
            variable_type_object = nested_variable_type_object
        else:
            cip_datatype_instance.variable_type_name = str(variable_type_object.variable_type_name, 'utf-8')
        cip_datatype_instance.size = variable_type_object.size_in_memory
        nest_id = variable_type_object.nesting_variable_type_instance_id
        member_instance_id = int.from_bytes(nest_id, 'little')
        while member_instance_id != 0:
            variable_type_object_reply = await self._get_variable_type_object(member_instance_id)
            member_cip_datatype_instance = await self._get_member_instance(member_instance_id, variable_type_object_reply)
            if type(member_cip_datatype_instance) == CIPStructure:
                member_cip_datatype_instance.callback = cip_datatype_instance.from_value
                member_cip_datatype_instance.callback_arg = cip_datatype_instance
                if variable_type_object.number_of_members == 0:
                    pass
            member_name = str(variable_type_object_reply.variable_type_name, 'utf-8')
            cip_datatype_instance.members[member_name] = member_cip_datatype_instance
            member_instance_id = \
                int.from_bytes(variable_type_object_reply.next_instance_id, 'little')
        return cip_datatype_instance

    async def _array_instance_from_variable_name(self, variable_name: str) -> CIPArray:
        """
        This method builds an array from information obtained from get_attribute_all on the
        variable_request_path
        :param variable_name:
        :return:
        """
        cip_array_instance = CIPArray()
        request_path = variable_request_path_segment(variable_name)
        response = await self.connected_cip_dispatcher.get_attribute_all_service(request_path)
        # Not actually a VariableObjectReply, but the data aligns the same
        array_attributes_all_reply = VariableObjectReply(response.bytes)
        instance_id = int.from_bytes(array_attributes_all_reply.variable_type_instance_id, 'little')
        cip_array_instance.instance_id = instance_id
        if array_attributes_all_reply.cip_data_type_of_array == CIPStructure.data_type_code():
            array_member_instance = await self._get_member_instance(instance_id)
        elif array_attributes_all_reply.cip_data_type_of_array == CIPAbbreviatedStructure.data_type_code():
            array_member_instance = await self._get_member_instance(instance_id)
        elif array_attributes_all_reply.cip_data_type_of_array == CIPString.data_type_code():
            array_member_instance = CIPString()
            array_member_instance.size = array_attributes_all_reply.size
        elif array_attributes_all_reply.cip_data_type_of_array == CIPArray.data_type_code():
            array_member_instance = await self._get_member_instance(instance_id)
        else:
            array_member_instance = \
                self.connected_cip_dispatcher.data_type_dictionary.get(
                    array_attributes_all_reply.cip_data_type_of_array)()

        cip_array_instance.from_instance(array_member_instance,
                                         array_attributes_all_reply.size,
                                         array_attributes_all_reply.array_dimension,
                                         array_attributes_all_reply.number_of_elements,
                                         array_attributes_all_reply.start_array_elements)
        cip_array_instance.variable_name = variable_name
        return cip_array_instance

    async def _array_instance_from_variable_type_object(self, variable_type_object: VariableTypeObjectReply) -> CIPArray:
        """
        This method builds an array from information obtained from get_attribute_all on the
        variable_request_path
        :param variable_type_object:
        :return:
        """
        cip_array_instance = CIPArray()
        # Not actually a VariableObjectReply, but the data aligns the same
        instance_id = variable_type_object.nesting_variable_type_instance_id
        if type(instance_id) is bytes:
            instance_id = int.from_bytes(instance_id, 'little', signed=False)
        cip_array_instance.instance_id = instance_id
        if variable_type_object.cip_data_type_of_array == CIPStructure.data_type_code():
            array_member_instance = await self._get_member_instance(instance_id)
        elif variable_type_object.cip_data_type_of_array == CIPAbbreviatedStructure.data_type_code():
            array_member_instance = await self._get_member_instance(instance_id)
        elif variable_type_object.cip_data_type_of_array == CIPString.data_type_code():
            cip_data_type_instance = await self._string_instance_from_variable_type_object(variable_type_object)
            cip_data_type_instance.size = variable_type_object.size_in_memory
            array_member_instance = cip_data_type_instance
        elif variable_type_object.cip_data_type_of_array == CIPArray.data_type_code():
            array_member_instance = await self._get_member_instance(instance_id)
        else:
            array_member_instance = \
                self.connected_cip_dispatcher.data_type_dictionary.get(
                    variable_type_object.cip_data_type_of_array)()
        cip_array_instance.from_instance(array_member_instance,
                                         variable_type_object.size,
                                         variable_type_object.array_dimension,
                                         variable_type_object.number_of_elements,
                                         variable_type_object.start_array_elements)
        return cip_array_instance

    @staticmethod
    async def _string_instance_from_variable_type_object(
            variable_object: VariableTypeObjectReply) -> CIPString:
        cip_string = CIPString()
        cip_string.size = variable_object.size
        return cip_string

    async def _get_instance_from_variable_name(self, variable_name: str):
        """
        This method returns the correct type of CIP Datatype instance based on the associated datatype
        that is discovered by calling the get_attribute_all CIP Service on the variable's request path.
        :param variable_name:
        :return:
        """
        request_path = variable_request_path_segment(variable_name)
        data_type_code = b''
        response = None
        # Split on square brackets and period characters to get tokens of derived data types
        derived_data_type_variable_name_split = list(filter(None, re.split(r'[\[\].]', variable_name)))
        try:
            response = await self.connected_cip_dispatcher.get_attribute_all_service(request_path)
            data_type_code = response.reply_data[4:5]
        except CIPException as e:
            """ 
            This CIP command can have an exception when it is a derived data type, and should be passed along
            to be processed in the else statement below, or it doesn't exist and should reraise the exception
            """
            if len(derived_data_type_variable_name_split) > 1:
                pass
            else:
                raise e
        if data_type_code == CIPStructure.data_type_code():
            variable_type_object_instance_id = int.from_bytes(response.reply_data[8:12], 'little')
            variable_type_object = await self._get_variable_type_object(variable_type_object_instance_id)
            cip_data_type_instance = await self._structure_instance_from_variable_type_object(variable_type_object)
        elif data_type_code == CIPAbbreviatedStructure.data_type_code():
            variable_type_object_instance_id = int.from_bytes(response.reply_data[8:12], 'little')
            variable_type_object = await self._get_variable_type_object(variable_type_object_instance_id)
            cip_data_type_instance = await self._structure_instance_from_variable_type_object(variable_type_object)
        elif data_type_code == CIPString.data_type_code():
            variable_type_object_instance_id = int.from_bytes(response.reply_data[8:12], 'little')
            variable_type_object = await self._get_variable_type_object(variable_type_object_instance_id)
            cip_data_type_instance = await self._string_instance_from_variable_type_object(variable_type_object)
            cip_data_type_instance.size = int.from_bytes(response.reply_data[0:2], 'little')
        elif data_type_code == CIPArray.data_type_code():
            cip_data_type_instance = await self._array_instance_from_variable_name(variable_name)
        elif data_type_code != b'':
            cip_data_type_instance = self.connected_cip_dispatcher.data_type_dictionary.get(data_type_code)()
        else:
            super_instance = await self._get_instance_from_variable_name(derived_data_type_variable_name_split[0])
            for token in derived_data_type_variable_name_split[1:]:
                if token.isnumeric():
                    if hasattr(super_instance, 'local_cip_data_type_object'):
                        super_instance = super_instance.local_cip_data_type_object
                else:
                    super_instance = super_instance[token]
            cip_data_type_instance = super_instance
        cip_data_type_instance.variable_name = str(variable_name)
        self.connected_cip_dispatcher.variables.update({variable_name: cip_data_type_instance})
        if variable_name[0:1] == '_':
            self.connected_cip_dispatcher.system_variables.update({variable_name: cip_data_type_instance})
        else:
            self.connected_cip_dispatcher.user_variables.update({variable_name: cip_data_type_instance})
        return cip_data_type_instance

    async def _get_member_instance(
            self, member_instance_id: int, variable_type_object_reply: VariableTypeObjectReply=None) -> CIPDataType:
        """
        This method returns a CIP datatype instance from a member ID. This is how the driver can
        build instances of derived data types like structures and arrays of structures
        :param member_instance_id:
        :param variable_type_object_reply:
        :return:
        """
        if variable_type_object_reply is None:
            variable_type_object_reply = await self._get_variable_type_object(member_instance_id)

        if variable_type_object_reply.cip_data_type == CIPStructure.data_type_code():
            return await self._structure_instance_from_variable_type_object(variable_type_object_reply)
        elif variable_type_object_reply.cip_data_type == CIPAbbreviatedStructure.data_type_code():
            return await self._structure_instance_from_variable_type_object(variable_type_object_reply)
        elif variable_type_object_reply.cip_data_type == CIPString.data_type_code():
            return await self._string_instance_from_variable_type_object(variable_type_object_reply)
        elif variable_type_object_reply.cip_data_type == CIPArray.data_type_code():
            return await self._array_instance_from_variable_type_object(variable_type_object_reply)
        else:
            return self.connected_cip_dispatcher.data_type_dictionary.get(variable_type_object_reply.cip_data_type)()

    async def _array_instance_from_variable_object(
            self, variable_object: (VariableObjectReply, VariableTypeObjectReply)) -> CIPArray:
        cip_array_instance = CIPArray()
        variable_type_instance_id = b'\x00\x00\x00\x00'
        if isinstance(variable_object, VariableObjectReply):
            variable_type_instance_id = variable_object.variable_type_instance_id
        if variable_type_instance_id == b'\x00\x00\x00\x00':
            cip_array_instance.from_items(variable_object.cip_data_type_of_array,
                                          variable_object.size,
                                          variable_object.array_dimension,
                                          variable_object.number_of_elements,
                                          variable_object.start_array_elements)
        else:
            member_instance_id = int.from_bytes(variable_object.variable_type_instance_id, 'little')
            member_instance = await self._get_member_instance(member_instance_id)
            cip_array_instance = member_instance
        return cip_array_instance

    async def read_variable(self, variable_name: str) -> CIPDataType:
        """
        This method will read the variable name from the controller and return it in the corresponding
        Python datatype
        :rtype: object
        :param variable_name:
        :return:
        """
        request_path = variable_request_path_segment(variable_name)
        cip_data_type_instance = self.connected_cip_dispatcher.variables.get(variable_name)
        if cip_data_type_instance is None:
            cip_data_type_instance = await self._get_instance_from_variable_name(variable_name)
        if isinstance(cip_data_type_instance, (CIPString, CIPArray, CIPStructure, CIPAbbreviatedStructure)):
            cip_data_type_instance.variable_name = variable_name
            if (isinstance(cip_data_type_instance, CIPArray) and
                    isinstance(cip_data_type_instance.local_cip_data_type_object, CIPString)):
                list_of_strings = []
                for index in range(cip_data_type_instance.number_of_elements[0]):
                    variable_response = await self.read_variable(f'{variable_name}[{index}]')
                    list_of_strings.append(variable_response.value())
                cip_data_type_instance.from_value(list_of_strings)
                return cip_data_type_instance
            else:
                return await self._multi_message_variable_read(cip_data_type_instance)
        else:
            response = await self.connected_cip_dispatcher.read_tag_service(request_path)
            cip_data_type_instance.from_bytes(response.reply_data)
            cip_data_type_instance.variable_name = variable_name
            return cip_data_type_instance.value()

    async def write_variable(self, variable_name: str, data):
        """
        This method takes a variable name and formats the Python datatype into the correct CIP datatype
        and writes it to the controller
        :param variable_name:
        :param data:
        :return:
        """
        request_path = variable_request_path_segment(variable_name)
        cip_data_type_instance = self.connected_cip_dispatcher.variables.get(variable_name)
        if cip_data_type_instance is None:
            cip_data_type_instance = await self._get_instance_from_variable_name(variable_name)
        cip_data_type_instance.from_value(data)
        if cip_data_type_instance is None:
            """Should be an exception"""
            pass
        elif isinstance(cip_data_type_instance, (CIPString, CIPArray, CIPStructure, CIPAbbreviatedStructure)):
            if (isinstance(cip_data_type_instance, CIPArray) and
                    isinstance(cip_data_type_instance.local_cip_data_type_object, CIPString)):
                list_of_strings = data
                for index in range(cip_data_type_instance.number_of_elements[0]):
                    await self.write_variable(f'{variable_name}[{index}]', data[index])
                return list_of_strings
            else:
                await self._multi_message_variable_write(cip_data_type_instance)
        else:
            request_data = CIPCommonFormat(cip_data_type_instance.data_type_code(), data=cip_data_type_instance.data)
            await self.connected_cip_dispatcher.write_tag_service(request_path, request_data)

    async def verified_write_variable(self, variable_name: str, data, retry_count: int = 2):
        remaining_retry = retry_count
        temp_data = None
        while remaining_retry >= 0 and temp_data != data:
            await self.write_variable(variable_name, data)
            temp_data = await self.read_variable(variable_name)
            remaining_retry += -1
        if remaining_retry < 0 and temp_data != data:
            logging.warning('Write Operation could not be completed within specified retry count')
            raise IOError('Write Operation could not be completed within specified retry count')

    async def _multi_message_variable_read(self, cip_datatype_object: CIPDataType, offset=0) -> CIPDataType:
        """
        This method is to read data that does not fit into a single CIP message
        :param cip_datatype_object:
        :param offset:
        :return:
        """
        max_read_size = self.MAXIMUM_LENGTH - 8
        data = b''
        while offset < cip_datatype_object.size:
            if cip_datatype_object.size - offset > max_read_size:
                read_size = max_read_size
            else:
                read_size = cip_datatype_object.size - offset
            response = await self._simple_data_segment_read(cip_datatype_object, offset, read_size)
            reply_bytes = response.reply_data
            cip_common_format = CIPCommonFormat()
            cip_common_format.from_bytes(reply_bytes)
            if isinstance(cip_datatype_object, CIPString):
                # First two characters of the string seem to be how many characters were read
                data = data + cip_common_format.data[2:]
            elif isinstance(cip_datatype_object, CIPStructure):
                cip_datatype_object.crc_code = cip_common_format.additional_info
                data = data + cip_common_format.data
            else:
                data = data + cip_common_format.data
            offset = offset + max_read_size
        cip_datatype_object.data = data
        # cip_datatype_object.size = len(data) # Removed Why did it exist? If weird stuff breaks revisit
        cip_datatype_object.value()
        return cip_datatype_object

    async def _multi_message_variable_write(self, cip_datatype_object: CIPDataType, offset=0) -> CIPReply:
        """
        This method is to write data that does not fit into a single CIP message
        :param cip_datatype_object:
        :param offset:
        :return:
        """
        max_write_size = 400
        response = CIPReply
        while offset < cip_datatype_object.size:
            if cip_datatype_object.size - offset > max_write_size:
                write_size = max_write_size
            else:
                write_size = cip_datatype_object.size - offset
            response = await self._simple_data_segment_write(
                cip_datatype_object, offset, write_size,
                cip_datatype_object.data[offset:offset + write_size])
            offset = offset + max_write_size
        return response

    async def _simple_data_segment_read(self, cip_datatype_object: CIPDataType, offset, read_size) -> CIPReply:
        """
        This method formats as request path to be used with simple_data_segment reading operations
        :param cip_datatype_object:
        :param offset:
        :param read_size:
        :return:
        """
        request_path = variable_request_path_segment(cip_datatype_object.variable_name)
        simple_data_request_path = SimpleDataSegmentRequest(offset, read_size)
        request_path = request_path + simple_data_request_path.bytes()
        response = await self.connected_cip_dispatcher.read_tag_service(request_path)
        return response

    async def _simple_data_segment_write(self, cip_datatype_object: CIPDataType, offset, write_size, data) -> CIPReply:
        """
        This method formats as request path to be used with simple_data_segment writing operations
        :param cip_datatype_object:
        :param offset:
        :param write_size:
        :param data:
        :return:
        """
        request_path = variable_request_path_segment(cip_datatype_object.variable_name)
        simple_data_request_path = SimpleDataSegmentRequest(offset, write_size)
        request_path = request_path + simple_data_request_path.bytes()
        response = None
        if cip_datatype_object.data_type_code() == CIPString.data_type_code():
            data = struct.pack("<H", len(data)) + data
            request_data = CIPCommonFormat(cip_datatype_object.data_type_code(), data=data)
            response = await self.connected_cip_dispatcher.write_tag_service(request_path, request_data)
        elif cip_datatype_object.data_type_code() == CIPArray.data_type_code():
            if cip_datatype_object.array_data_type == CIPStructure.data_type_code():
                structure_variable_type_object = \
                    await self._get_variable_type_object(cip_datatype_object.instance_id)
                crc_code = structure_variable_type_object.crc_code.to_bytes(2, 'little')
                request_data = CIPCommonFormat(CIPAbbreviatedStructure.data_type_code(), additional_info_length=2,
                                               additional_info=crc_code,
                                               data=data)
            else:
                request_data = CIPCommonFormat(cip_datatype_object.array_data_type, data=data)
            response = await self.connected_cip_dispatcher.write_tag_service(request_path, request_data)
        elif cip_datatype_object.data_type_code() == CIPStructure.data_type_code():
            request_data = CIPCommonFormat(CIPAbbreviatedStructure.data_type_code(), additional_info_length=2,
                                           additional_info=cip_datatype_object.crc_code, data=data)
            response = await self.connected_cip_dispatcher.write_tag_service(request_path, request_data)
        return response

    async def _get_variable_object(self, instance_id: int) -> VariableObjectReply:
        """
        Omron specific CIP class that is used to describe variables. This contains critical information if the
        variable is an array type
        :param instance_id:
        :return:
        """
        request_path = address_request_path_segment(
            class_id=b'\x6b', instance_id=instance_id.to_bytes(2, 'little'))
        variable_object_reply = VariableObjectReply(
            self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
        return variable_object_reply

    async def _get_variable_type_object(self, instance_id: int) -> VariableTypeObjectReply:
        """
        Omron specific CIP class that is used to describe variable types. This is where derived data types
        will have their member definitions
        :param instance_id:
        :return:
        """
        request_path = address_request_path_segment(
            class_id=b'\x6c', instance_id=instance_id.to_bytes(2, 'little'))
        variable_type_object_reply_value = await self.connected_cip_dispatcher.get_attribute_all_service(request_path)
        variable_type_object_reply = VariableTypeObjectReply(variable_type_object_reply_value.bytes)
        return variable_type_object_reply

    async def _get_number_of_derived_data_types(self) -> int:
        request_path = eip.address_request_path_segment(class_id=b'\x6c', instance_id=int(0).to_bytes(2, 'little'))
        reply = self.connected_cip_dispatcher.get_attribute_all_service(request_path)
        max_instance = struct.unpack("<H", reply.reply_data[2:4])[0]
        return max_instance

    async def _get_variable_list(self):
        """
                Omron specific method for creating a list of variables that are published in the controller.
                :return:
                """
        tag_list = self._get_system_variable_list() + self._get_user_variable_list()
        return tag_list

    async def _get_user_variable_list(self):
        return self._get_variable_list_subset(True)

    async def _get_system_variable_list(self):
        return self._get_variable_list_subset(False)

    async def _get_instance_list_subset(self, user_defined: bool):
        all_received = False
        current_instance = 1
        instance_list = []
        while not all_received:
            reply = await self.connected_cip_dispatcher.get_instance_list(current_instance, 100, user_defined)
            instance_count = struct.unpack('<H', reply.reply_data[0:2])[0]
            unprocessed_instances = instance_count
            if reply.reply_data[2] == 0 or instance_count == 0:
                all_received = True
            instances = reply.reply_data[4:]
            while unprocessed_instances != 0:
                data_length = struct.unpack('<H', instances[4:6])[0]
                instance = InstanceIDAttributes(instances[4:4 + data_length + 2])
                instance_list.append(instance)
                instances = instances[4 + data_length + 2:]
                unprocessed_instances -= 1
            current_instance = current_instance + instance_count
        return instance_list

    async def _get_variable_list_subset(self, user_defined: bool):
        # all_received = False
        instances = await self._get_instance_list_subset(user_defined)
        tag_list = []
        for instance in instances:
            tag_list.append(instance.tag_name())
        return tag_list

    async def _get_number_of_variables(self) -> int:
        """
        Omron specific method to find number of variables from Tag Name Server
        :return:
        """
        request_path = address_request_path_segment(b'\x6a', b'\x00\x00')
        reply = self.connected_cip_dispatcher.get_attribute_all_service(request_path)
        return int.from_bytes(reply.reply_data[2:4], 'little')


class EIPConnectionStatus:
    def __init__(self):
        self.connected = False
        self.connecting = False
        self.reconnecting = False
        self._has_session = False
        self.persist_session = False
        self._session_status_observers = []
        self.keep_alive = False
        self.keep_alive_running = False

    def bind_to_session_status(self, callback):
        """
        Observer Pattern: Allow widgets to bind a callback function that will act on reestablished session
        :param callback:
        :return:
        """
        self._session_status_observers.append(callback)

    @property
    def has_session(self):
        return self._has_session

    @has_session.setter
    def has_session(self, value):
        self._has_session = value
        for callback in self._session_status_observers:
            callback()


class MonitoredVariable:
    instance = {}

    def __new__(cls, dispatcher: "NSeriesThreadDispatcher", variable_name,
                refresh_time: float = 0.05, *args, **kwargs):
        unique_key = str(dispatcher) + variable_name
        if unique_key in MonitoredVariable.instance:
            monitored_object = MonitoredVariable.instance[unique_key]
            if monitored_object.refresh_time > refresh_time:
                monitored_object.refresh_time = refresh_time
            return monitored_object
        else:
            return super(MonitoredVariable, cls).__new__(cls)

    def __init__(self, dispatcher: "NSeriesThreadDispatcher", variable_name,
                 refresh_time: float = 0.05, observers=[]):
        self.variable_name = variable_name
        self.unique_key = str(dispatcher) + self.variable_name
        self.dispatcher = dispatcher
        self.monitored_variable_observers = observers
        self.refresh_time = refresh_time
        self.refresh_timer = threading.Timer(self.refresh_time, self.update)
        self._value = None
        MonitoredVariable.instance[self.unique_key] = self
        self.update()

    def bind_to_value(self, callback):
        """
        Observer Pattern: Allow widgets to bind a callback function that will act on value change
        :param callback:
        :return:
        """
        self.monitored_variable_observers.append(callback)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self.dispatcher.verified_write_variable(self.variable_name, value)
        self._value = value
        for callback in self.monitored_variable_observers:
            callback()

    def update(self):
        temp_value = self._value
        self._value = self.dispatcher.read_variable(self.variable_name)
        if temp_value != self._value:
            for callback in self.monitored_variable_observers:
                callback()
        self.refresh_timer = threading.Timer(self.refresh_time, self.update)
        self.refresh_timer.daemon = True
        self.refresh_timer.start()

    def start(self):
        self.refresh_timer.start()

    def cancel(self):
        self.refresh_timer.cancel()


class NSeriesThreadDispatcher:
    def __init__(self, host: str = None, connection_timeout: float = None,
                 retry_time: float = 1.0, max_attempts: int = None):
        self._instance = NSeries(host, connection_timeout)
        self._host = None
        self.message_timeout = 0.5
        self.executor = None
        self.futures = []
        self.status_integer = 0
        self.services = None
        self.monitored_variable_dictionary = {}
        self.connection_status = EIPConnectionStatus()
        if host is not None:
            self.connect_explicit(host, connection_timeout, retry_time, max_attempts)
            # self.register_session(retry_time)

    def add_monitored_variable(self, monitored_variable: MonitoredVariable):
        self.monitored_variable_dictionary[monitored_variable.variable_name] = monitored_variable

    def _execute_eip_command(self, command, *args, **kwargs):
        if self.connection_status.connected:
            future = self.executor.submit(command, *args, **kwargs)
            try:
                result = future.result()
                return result
            except socket.error as err:
                print(f'{err} occurred in _execute_eip_command')
                self.connection_status.connected = False
                self.connection_status.keep_alive_running = False
                self.executor.shutdown()
                self._reconnect()
        else:
            self._reconnect()

    def retry_command(self, command, retry_time: float = 0.05, *args, **kwargs):
        reply = command(*args, **kwargs)
        if not reply:
            delay = threading.Timer(retry_time, self.retry_command, command,
                                    *args, **dict(**kwargs, retry_time=retry_time))
            delay.daemon = True
            delay.start()
        return reply

    def _reconnect(self, already_executing=False):
        if not already_executing and not self.connection_status.reconnecting:
            self.connection_status.reconnecting = True
            self._reconnect(already_executing=True)
        elif self.connection_status.reconnecting and already_executing:
            if not self.connection_status.connected:
                try:
                    self.close_explicit()
                except Exception as err:
                    print(f'{err} occurred in close_explicit')
                self.connect_explicit(self._host)
                self._reconnect(already_executing=True)
            if (self.connection_status.persist_session
                    and self.connection_status.connected
                    and not self.connection_status.has_session):
                self.register_session()
                self._reconnect(already_executing=True)
            elif not self.connection_status.persist_session and not self.connection_status.keep_alive:
                self.connection_status.reconnecting = False
            if self.connection_status.keep_alive and self.connection_status.has_session:
                self.start_keep_alive()
            elif not self.connection_status.keep_alive:
                self.connection_status.reconnecting = False

    def start_keep_alive(self, keep_alive_time: float = 0.05):
        self.connection_status.keep_alive = True
        if (self.connection_status.connected
                and self.connection_status.keep_alive
                and self.connection_status.has_session):
            self.connection_status._keep_alive_running = True
            self.services = self._execute_eip_command(self._instance.connected_cip_dispatcher.list_services, '')
            delay = threading.Timer(keep_alive_time, self.start_keep_alive, [keep_alive_time])
            delay.daemon = True
            delay.start()

    def connect_explicit(self, host: str, connection_timeout: float = None,
                         retry_time: float = 1.0, max_attempts: int = None):
        self.connection_status.connecting = True
        if not self.connection_status.connected:
            self._host = host
            try:
                self._instance.connect_explicit(host=self._host, connection_timeout=connection_timeout)
                self.executor = concurrent.futures.ThreadPoolExecutor()
                self.connection_status.connected = self._instance.connected_cip_dispatcher.is_connected_explicit
                self.connection_status.connecting = False
            except socket.error as err:
                print(f"Failed to connect, trying again in {retry_time} seconds")
                if max_attempts is not None:
                    max_attempts -= 1
                    print(f'Will retry {max_attempts} more times')
                if max_attempts != 0:
                    time.sleep(retry_time)
                    self.connect_explicit(host, connection_timeout, retry_time, max_attempts)
                    # delay = threading.Timer(
                    #     retry_time, self.connect_explicit, [host, connection_timeout, retry_time, max_attempts])
                    # delay.start()
                else:
                    raise err

    def register_session(self, retry_time: float = 1.0):
        self._instance.register_session()
        session_id = self._instance.connected_cip_dispatcher.session_handle_id
        if session_id != b'\x00\x00\x00\x00\x00\x00\x00\x00':
            self.connection_status.has_session = True
        else:
            print(f"Failed to register session, trying again in {retry_time} seconds")
            delay = threading.Timer(retry_time, self.register_session, [retry_time])
            delay.start()
        self.connection_status.persist_session = True

    @property
    def connected_cip_dispatcher(self):
        return self._instance.connected_cip_dispatcher

    def unregister_session(self):
        self.connection_status.has_session = False
        self.connection_status.persist_session = False

    def update_variable_dictionary(self):
        self._instance.update_variable_dictionary()

    def variable_list(self):
        return self._instance.variable_list()

    def user_variable_list(self):
        return self._instance.user_variable_list()

    def system_variable_list(self):
        return self._instance.system_variable_list()

    def save_current_dictionary(self, file_name: str):
        self._instance.save_current_dictionary(file_name)

    def load_dictionary_file(self, file_name: str):
        self._instance.load_dictionary_file(file_name)

    def load_dictionary_file_if_present(self, file_name: str):
        self._instance.load_dictionary_file_if_present(file_name)

    def read_variable(self, variable_name: str):
        try:
            return self._execute_eip_command(self._instance.read_variable, variable_name)
        except struct.error as error:
            raise error

    def write_variable(self, variable_name: str, data):
        return self._execute_eip_command(self._instance.write_variable, variable_name, data)

    def verified_write_variable(self, variable_name: str, data):
        try:
            return self._execute_eip_command(self._instance.verified_write_variable, variable_name, data)
        except struct.error as error:
            raise error

    def close_explicit(self):
        self.connection_status.connected = False
        self.connection_status.has_session = False
        self.connection_status._keep_alive_running = False
        for monitored in self.monitored_variable_dictionary:
            self.monitored_variable_dictionary[monitored].cancel()
        time.sleep(0.01)
        self.executor.shutdown(wait=True)
        if self._instance.connected_cip_dispatcher.is_connected_explicit:
            self._instance.close_explicit()

    def __enter__(self):
        signal(SIGINT, self._sigint_handler)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_explicit()

    def _sigint_handler(self, signal_received, frame):
        print('Program interrupted with Ctrl+C')
        self.__exit__(None, None, None)
        sys.exit(0)



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import collections
import struct
import copy
from abc import ABC, abstractmethod
import binascii


def test_bit(int_type, offset):
    mask = 1 << offset
    return int_type & mask != 0


def set_bit(int_type, offset):
    mask = 1 << offset
    return int_type | mask


def clear_bit(int_type, offset):
    mask = ~(1 << offset)
    return int_type & mask


def toggle_bit(int_type, offset):
    mask = 1 << offset
    return int_type ^ mask


def flatten(x):
    if isinstance(x, collections.Iterable):
        return [a for i in x for a in flatten(i)]
    else:
        return [x]


class CIPDataType(ABC):
    """
    Abstract Base Class for CIP Data Types. The subclasses will be added to a data type dictionary and
    based on the data type code, the value method will provide an appropriate representation of the
    data to the client program when it reads from a CIP device and  the from_value method will format
    python types to the proper byte representation to send to the CIP device on write operations
    """

    @staticmethod
    @abstractmethod
    def data_type_code() -> bytes:
        """pass"""

    def __init__(self):
        self._data_type_code = self.data_type_code()
        self.addition_info_length = 0
        self.additional_info = b''
        self.data = bytearray()
        self.size = len(self.data)
        self.instance_id = None
        self.variable_name = ''

    def __repr__(self):
        return str(self.value())

    @property
    def alignment(self) -> int:
        """
        Data types have a memory alignment that describe how they are stored in memory which needs to be
        accounted for when reading or writing structures. Most basic data types align to their size, but
        strings align to a single byte, arrays to the size of their members (except Booleans), and unions
        and structures to the largest alignment of their members. Those types will override this method
        :return:
        """
        return self.size

    def bytes(self):
        byte_value = self.data_type_code() + \
                     self.addition_info_length.to_bytes(1, 'little') + \
                     self.additional_info + \
                     self.data
        return byte_value

    def from_bytes(self, bytes_cip_data_type):
        self._data_type_code = bytes_cip_data_type[0:1]
        self.addition_info_length = int.from_bytes(bytes_cip_data_type[1:2], 'little')
        self.additional_info = bytes_cip_data_type[2:2 + self.addition_info_length]
        self.data = bytes_cip_data_type[2 + self.addition_info_length:]

    @abstractmethod
    def value(self):
        """pass Abstract Docstring to pass coverage"""

    @abstractmethod
    def from_value(self, value):
        """pass Abstract Docstring to pass coverage"""


def update_data_type_dictionary(data_type_dictionary):
    """
    This function creates a datatype dictionary from all the subclasses that inherit from the
    CIP datatype abstract base class

    :param data_type_dictionary:
    :return:
    """
    for sub_class in CIPDataType.__subclasses__():
        data_type_dictionary.update({sub_class.data_type_code(): sub_class})


def _get_class_data_type_code(data_type: bytes) -> CIPDataType | None:
    """
    This function will return a CIP datatype instance from a datatype code

    :param data_type:
    :return:
    """
    for sub_class in CIPDataType.__subclasses__():
        if sub_class.data_type_code() == data_type:
            return sub_class()
    return None


class CIPBoolean(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = struct.pack("<h", 0)
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc1'  # (bit)

    def value(self):
        value = int.from_bytes(self.data, 'little')
        return value != 0

    def from_value(self, value):
        self.data = struct.pack("<h", 1 if value else 0)


class CIPShortInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc2'  # (1-byte signed binary)

    def value(self):
        return struct.unpack("<b", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<b", value)


class CIPInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc3'  # (1-word signed binary)

    def value(self):
        return struct.unpack("<h", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<h", value)


class CIPDoubleInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc4'  # (2-word signed binary)

    def value(self):
        return struct.unpack("<l", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<l", value)


class CIPLongInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc5'  # (1-byte signed binary)

    def value(self):
        return struct.unpack("<q", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<q", value)


class CIPUnsignedShortInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc6'  # (1-byte unsigned binary)

    def value(self):
        return struct.unpack("<B", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<B", value)


class CIPUnsignedInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc7'  # (1-word unsigned binary)

    def value(self):
        return struct.unpack("<H", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<H", value)


class CIPUnsignedDoubleInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc8'  # (2-word unsigned binary)

    def value(self):
        return struct.unpack("<L", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<L", value)


class CIPUnsignedLongInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc9'  # (4-word unsigned binary)

    def value(self):
        return struct.unpack("<Q", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<Q", value)


class CIPReal(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xca'  # (2-word floating point)

    def value(self):
        return struct.unpack("<f", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<f", value)


class CIPLongReal(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xcb'  # (4-word floating point)

    def value(self):
        return struct.unpack("<d", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<d", value)


class CIPString(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = str("").encode('utf-8')

    def __eq__(self, other):
        return other == str(self)

    @staticmethod
    def data_type_code():
        return b'\xd0'  #

    @property
    def alignment(self) -> int:
        return 0

    def value(self):
        return str(self.data, 'utf-8').rsplit('\0')[0]

    def from_value(self, value):
        byte_value = value.encode('utf-8')
        length_difference = self.size - len(byte_value)
        byte_value += length_difference * b'\x00'
        self.data = byte_value
        self.size = len(self.data)


class CIPByte(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xd1'

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPWord(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xd2'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPDoubleWord(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xd3'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPLongWord(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xd4'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPTime(CIPDataType):
    def __init__(self):
        super().__init__()

    @staticmethod
    def data_type_code():
        return b'\xdb'  # (8-byte data)

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPAbbreviatedStructure(CIPDataType):
    def __init__(self):
        super().__init__()

    @staticmethod
    def data_type_code():
        return b'\xa0'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPStructure(CIPDataType):
    def __init__(self):
        super().__init__()
        self.variable_type_name = ''
        self.members = {}
        self._alignment = 0
        self.crc_code = b''
        self.callback = None
        self.callback_arg = None

    @staticmethod
    def data_type_code():
        return b'\xa2'  #

    @property
    def alignment(self) -> int:
        return self._alignment

    @alignment.setter
    def alignment(self, new_alignment):
        self._alignment = new_alignment

    def __setitem__(self, key, value):
        current_type = self.members.get(key)
        current_type.from_value(value)
        self.members[key] = current_type
        self.from_value(self)

    def __getitem__(self, item):
        return self.members[item]

    def __repr__(self):
        return '%s { type: %s | members: %s }' % (self.variable_name, self.variable_type_name, self.members)

    def add_member(self, member_name: str, member: CIPDataType):
        self.members[member_name] = member
        if member.alignment > self.alignment:
            self.alignment = member.alignment

    def value(self):
        offset = 0
        prev_member_value_type = None
        structure_data = self.data
        for member in self.members:
            member_value = self.members[member]
            if member_value.alignment != 0 and offset % member_value.alignment != 0:
                offset += (member_value.alignment - offset % member_value.alignment)
            elif prev_member_value_type != type(member_value) and offset % 2 != 0:
                # Make start offset even on member type change
                offset += 1
            end_byte = offset + member_value.size
            member_value.data = structure_data[offset:end_byte]
            # Call value to handle nested structures
            member_value.value()
            prev_member_value_type = type(member_value)
            offset = end_byte
        return self

    def from_value(self, value):
        # ToDo lookup data check
        offset = 0
        self.crc_code = value.crc_code
        prev_member_value_type = None
        mutable_data = bytearray(self.data)
        for member_key in value.members:
            member = value.members.get(member_key)
            if member.alignment != 0 and offset % member.alignment != 0:
                offset += (member.alignment - offset % member.alignment)
            elif prev_member_value_type != type(member) and offset % 2 != 0:
                # Make start offset even on member type change
                offset += 1
            end_byte = offset + member.size
            mutable_data[offset:offset+member.size] = member.data
            prev_member_value_type = type(member)
            offset = end_byte
        self.data = bytes(mutable_data)
        if self.callback is not None:
            self.callback(self.callback_arg)
        self.value()


class CIPArray(CIPDataType):
    # ToDo arrays of Boolean data types are a special case as they can pack 16 bits in their alignment
    def __init__(self):
        super().__init__()
        self.array_data_type = b''
        self.array_data_type_size = 0
        self.member_instance_id = None
        self.array_dimensions = 0
        self.number_of_elements = []
        self.start_array_elements = []
        self.size = 0
        self.local_cip_data_type_object = None
        self.list_representation = []
        self._alignment = 0
        self.data = b''

    def __getitem__(self, i):
        return self.list_representation[i]

    def __setitem__(self, i, value):
        self.list_representation[i] = value

    def __len__(self):
        return len(self.list_representation)

    def __eq__(self, other):
        return other == self.list_representation

    @property
    def alignment(self) -> int:
        return self._alignment

    def _get_size(self) -> int:
        """
        Determine the amount of data required to send and receive data,
        handling the special case of Boolean array packing
        :return:
        """
        total_size = 1
        last_number = None
        for number in self.number_of_elements:
            total_size = total_size * number
            last_number = number
        add_bytes = 0
        if self.array_data_type == b'\xc1':
            if last_number % 16 != 0:
                add_bytes = 2
            total_size = total_size // 8 + add_bytes
        else:
            total_size = total_size * self.array_data_type_size
        return total_size

    def from_items(self, array_data_type, array_data_size, array_dimensions, number_of_elements, start_array_elements):
        self.array_data_type = array_data_type
        self.array_data_type_size = array_data_size
        self.array_dimensions = array_dimensions
        self.number_of_elements = number_of_elements
        self.start_array_elements = start_array_elements
        self.size = self._get_size()
        self.local_cip_data_type_object = _get_class_data_type_code(self.array_data_type)
        if self.data == b'':
            self.data = self.size * b'\x00'
        self._alignment = self.local_cip_data_type_object.alignment

    def from_instance(self, cip_instance: CIPDataType, array_data_size,
                      array_dimensions, number_of_elements, start_array_elements):
        self.array_data_type = cip_instance.data_type_code()
        self.member_instance_id = cip_instance.instance_id
        self.array_data_type_size = array_data_size
        self.array_dimensions = array_dimensions
        self.number_of_elements = number_of_elements
        self.start_array_elements = start_array_elements
        self.size = self._get_size()
        self.local_cip_data_type_object = cip_instance
        if self.data == b'':
            self.data = self.size * b'\x00'
        self._alignment = self.local_cip_data_type_object.alignment

    def _recursive_data_to_array(self, dimension: int = 0, position: int = 0):
        """
        Takes the dimension value of the array and converts the data bytes
        returned into lists nested to the dimension depth
        """
        local_list = []
        elements_per_index = 1
        for element in range(dimension + 1, self.array_dimensions):
            elements_per_index = elements_per_index * self.number_of_elements[element]
        if dimension == self.array_dimensions - 1:
            temp_array = []
            for element in range(self.number_of_elements[dimension]):
                if self.array_data_type == b'\xc1':
                    start_bit = (position + element)
                    if test_bit(int.from_bytes(self.data, 'little'), start_bit):
                        integer_bool = 1
                    else:
                        integer_bool = 0
                    self.local_cip_data_type_object.data = integer_bool.to_bytes(2, 'little')
                else:
                    start_bytes = (position + element) * self.array_data_type_size
                    self.local_cip_data_type_object.data = \
                        self.data[start_bytes:start_bytes + self.array_data_type_size]
                temp_value = copy.deepcopy(self.local_cip_data_type_object)
                temp_array.append(temp_value.value())
            return temp_array
        else:
            for element in range(self.number_of_elements[dimension]):
                local_list.append(
                    self._recursive_data_to_array(dimension + 1, position + elements_per_index * element))
            return local_list

    def _recursive_array_to_data(self, dimension: int, list_data):
        """
        This method takes the dimension value of the array object and recursively converts the list into byte data
        """

        if dimension == 1:
            data = b''
            for element in list_data:
                self.local_cip_data_type_object.from_value(element)
                data += self.local_cip_data_type_object.data
            return data
        else:
            data = b''
            for sub_list in list_data:
                data += self._recursive_array_to_data(dimension - 1, sub_list)
            return data

    @staticmethod
    def data_type_code():
        return b'\xa3'  # (1-byte signed binary) signed char

    def value(self):
        self.list_representation = self._recursive_data_to_array()
        return self.list_representation

    def from_value(self, value):
        if self.array_data_type == b'\xc1':
            flat_list = flatten(self.list_representation)
            data_int = 0
            offset = 0
            for item in flat_list:
                if item:
                    data_int = set_bit(data_int, offset)
                else:
                    data_int = clear_bit(data_int, offset)
                offset = offset + 1
            self.data = data_int.to_bytes(self.size, 'little')
        else:
            self.data = self._recursive_array_to_data(self.array_dimensions, value)
        # Call value to run the recursive data to array
        self.value()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aphyt",
    version="0.1.25",
    author="Joseph Ryan",
    author_email="jr@aphyt.com",
    description="A library to communicate with Omron NX and NJ PLC and motion controllers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aphyt/aphytcomm",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    package_data={"aphyt.tkmi": ["Transparent_Pixel.png"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
)



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import asyncio
from abc import ABC, abstractmethod
from aphyt.cip.cip import CIPReply, CIPRequest, AsyncCIPDispatcher, CIPDispatcher, CIPException
import struct
import binascii
import socket
import time
from typing import List, Tuple
from aphyt.cip import *


class DataAndAddressItem:
    """
    Data and address items store the data used in common packet format
    EIP-CIP-V2-1.0.pdf Table 2-7.2 – Data and Address item format
    """
    NULL_ADDRESS_ITEM = b'\x00\x00'
    CONNECTED_TRANSPORT_PACKET = b'\xb1\x00'
    UNCONNECTED_MESSAGE = b'\xb2\x00'
    LIST_SERVICES_RESPONSE = b'\x00\x01'
    SOCKADDR_INFO_ORIGINATOR_TO_TARGET = b'\x00\x80'
    SOCKADDR_INFO_TARGET_TO_ORIGINATOR = b'\x01\x80'
    SEQUENCED_ADDRESS_ITEM = b'\x02\x80'

    def __init__(self, type_id, data: bytes):
        self.type_id = type_id
        self.length = len(data).to_bytes(2, 'little')
        self.data = data

    def from_bytes(self, bytes_data_address_item: bytes):
        self.type_id = bytes_data_address_item[0:2]
        self.length = bytes_data_address_item[2:4]
        self.data = bytes_data_address_item[4:]

    def bytes(self):
        return self.type_id + self.length + self.data


class CommonPacketFormat:
    """
    Common Packet Format stores the Data and Address packets. Used in both requests and replies.
    """

    def __init__(self, packets: List[DataAndAddressItem]):
        self.packets = [DataAndAddressItem(DataAndAddressItem.NULL_ADDRESS_ITEM, b''),
                        DataAndAddressItem(DataAndAddressItem.NULL_ADDRESS_ITEM, b'')]
        if len(packets) < 1:
            pass
        elif len(packets) < 2:
            self.packets[1] = packets[0]
        else:
            self.packets = packets
        self.item_count = len(self.packets)
        self.packet_bytes = b''
        for packet in self.packets:
            self.packet_bytes += packet.bytes()

    def from_bytes(self, bytes_common_packet_format: bytes):
        self.item_count = int.from_bytes(bytes_common_packet_format[0:2], 'little')
        self.packet_bytes = bytes_common_packet_format[2:]
        packet_offset = 0
        packet_index = 0
        while packet_offset < len(self.packet_bytes):
            data_address_item_id = self.packet_bytes[packet_offset: packet_offset + 2]
            data_address_item_length = int.from_bytes(self.packet_bytes[packet_offset + 2: packet_offset + 4], 'little')
            data_address_item_data = self.packet_bytes[packet_offset + 4: packet_offset + data_address_item_length + 4]
            self.packets[packet_index] = DataAndAddressItem(data_address_item_id, data_address_item_data)
            packet_offset = packet_offset + data_address_item_length + 4
            packet_index = packet_index + 1

    def bytes(self):
        return self.item_count.to_bytes(2, 'little') + self.packet_bytes


class CommandSpecificData:
    """
    Command Specific Data is sent in an EIP send rr command
    """

    def __init__(self, interface_handle: bytes = b'\x00\x00\x00\x00',
                 timeout: bytes = b'\x08\x00',
                 encapsulated_packet: bytes = b''):
        self.interface_handle = interface_handle
        self.timeout = timeout
        self.encapsulated_packet = encapsulated_packet

    def from_bytes(self, bytes_command_specific_data: bytes):
        self.interface_handle = bytes_command_specific_data[0:4]
        self.timeout = bytes_command_specific_data[4:6]
        self.encapsulated_packet = bytes_command_specific_data[6:]

    def bytes(self):
        return self.interface_handle + self.timeout + self.encapsulated_packet


class EIPMessage:
    """
    Class for creating or parsing Ethernet/IP messages

    EIP Message::
        |-Command Specific Data
          |-Common Packet Format
            |-Data And Address Item
              |-CIP Message
                |-Route Path
    """

    def __init__(self, command=b'\x00\x00', command_data=b'', session_handle_id=b'\x00\x00\x00\x00',
                 status=b'\x00\x00\x00\x00', sender_context_data=b'\x00\x00\x00\x00\x00\x00\x00\x00',
                 command_options=b'\x00\x00\x00\x00'):
        """

        :param command:
        :param command_data:
        :param session_handle_id:
        :param status:
        :param sender_context_data:
        :param command_options:
        """
        self.command = command
        self.length = len(command_data).to_bytes(2, 'little')
        self.session_handle_id = session_handle_id
        self.status = status
        self.sender_context_data = sender_context_data
        self.command_options = command_options
        self.command_data = command_data

    def bytes(self) -> bytes:
        return \
            self.command + self.length + self.session_handle_id + self.status + \
            self.sender_context_data + self.command_options + self.command_data

    def from_bytes(self, eip_message_bytes: bytes):
        self.command = eip_message_bytes[0:2]
        self.length = eip_message_bytes[2:4]
        self.session_handle_id = eip_message_bytes[4:8]
        self.status = eip_message_bytes[8:12]
        self.sender_context_data = eip_message_bytes[12:20]
        self.command_options = eip_message_bytes[20:24]
        self.command_data = eip_message_bytes[24:]

    def context_integer(self):
        return struct.unpack('<Q', self.sender_context_data)[0]

    def set_context(self, value: int):
        self.sender_context_data = struct.pack('<Q', value)

    def total_length(self):
        return struct.unpack('<H', self.length)[0] + 24


class EIPDispatcher(ABC):
    explicit_message_port = 44818

    def __init__(self):
        super().__init__()
        self.socket = None
        self.eip_responses = {}
        self.message_number = 18446744073709551613 # Using a number close to the max, fail fast on rollover errors

    @abstractmethod
    def send_command(self, eip_command: EIPMessage, host: str) -> EIPMessage:
        pass

    def list_identity(self, host=''):
        """
        Used by an originator to locate possible targets
        :return:
        """
        eip_message = EIPMessage(b'\x63\x00')
        return self.send_command(eip_message, host).command_data

    def list_services(self, host=''):
        """
        Find which services a target supports
        :return:
        """
        eip_message = EIPMessage(b'\x04\x00')
        return self.send_command(eip_message, host).command_data

    def list_interfaces(self, host=''):
        """
        Used by an originator to identify possible non-CIP interfaces on the target
        :return:
        """
        eip_message = EIPMessage(b'\x64\x00')
        return self.send_command(eip_message, host).command_data


class AsyncEIPDispatcher(ABC):
    explicit_message_port = 44818

    def __init__(self):
        super().__init__()
        self.socket = None
        self.eip_responses = {}
        self.message_number = 18446744073709551613 # Using a number close to the max, fail fast on rollover errors

    @abstractmethod
    async def send_command(self, eip_command: EIPMessage, host: str) -> EIPMessage:
        pass

    async def list_identity(self, host=''):
        """
        Used by an originator to locate possible targets
        :return:
        """
        eip_message = EIPMessage(b'\x63\x00')
        return await self.send_command(eip_message, host).command_data

    async def list_services(self, host=''):
        """
        Find which services a target supports
        :return:
        """
        eip_message = EIPMessage(b'\x04\x00')
        return await self.send_command(eip_message, host).command_data

    async def list_interfaces(self, host=''):
        """
        Used by an originator to identify possible non-CIP interfaces on the target
        :return:
        """
        eip_message = EIPMessage(b'\x64\x00')
        return await self.send_command(eip_message, host).command_data


class AsyncEIPConnectedCommandMixin(AsyncEIPDispatcher):
    lock = asyncio.Lock()
    def __init__(self):
        super().__init__()
        self.explicit_message_socket = None
        self.stream_writer = None
        self.stream_reader = None
        self.session_handle_id = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.is_connected_explicit = False
        self.has_session_handle = False
        self.BUFFER_SIZE = 4096
        self.host = None

    async def get_response(self, eip_message: EIPMessage):
        if eip_message.context_integer() in self.eip_responses.keys():
            result = self.eip_responses.pop(eip_message.context_integer())
            return result
        else:
            self.stream_writer.write(eip_message.bytes())
            await self.stream_writer.drain()
            async with AsyncEIPConnectedCommandMixin.lock:
                received_data = await self.stream_reader.readexactly(n=24)
                received_data += await self.stream_reader.readexactly(n=struct.unpack('<H',received_data[2:4])[0])
            length = struct.unpack('<H',received_data[2:4])[0] + 24
            received_eip_message = EIPMessage()
            received_eip_message.from_bytes(received_data[0:length])
            # Below recreates the extraction of a CIP reply that normally occurs in execute_cip_command and
            # send_rr_data, but we want to check for x02 resource unavailable response
            reply_command_specific_data = CommandSpecificData()
            reply_command_specific_data.from_bytes(received_eip_message.command_data)
            reply_packet = CommonPacketFormat([])
            reply_packet.from_bytes(reply_command_specific_data.encapsulated_packet)
            response = reply_packet.packets[1].bytes()
            reply_data_and_address_item = DataAndAddressItem('', b'')
            reply_data_and_address_item.from_bytes(response)
            cip_reply = CIPReply(reply_data_and_address_item.data)
            if cip_reply.general_status == b'\x02':
                # Don't add the response to the eip_responses dictionary if it is "resource unavailable"
                # because we should try again because the resource was flooded
                pass
            else:
                self.eip_responses[received_eip_message.context_integer()] = received_eip_message
            try:
                response = await self.get_response(eip_message)
            except RecursionError as e:
                if cip_reply.general_status == b'\x02':
                    raise CIPException(b'\x02', b'')
                else:
                    raise e
            return response


    async def send_command(self, eip_command: EIPMessage, host) -> EIPMessage:
        """
        Used to send and receive Ethernet/IP messages
        :param host:
        :param eip_command:
        :return:
        """
        # received_eip_message = EIPMessage()
        if self.message_number == 18446744073709551616:
            self.message_number = 0
        eip_command.set_context(self.message_number)
        self.message_number = self.message_number + 1
        return await self.get_response(eip_command)

    async def connect_explicit(self, host, connection_timeout: float = None):
        """
        Create and explicit Ethernet/IP connection
        :param connection_timeout:
        :param host:
        """
        try:
            self.stream_reader, self.stream_writer = (
                await asyncio.open_connection(host, self.explicit_message_port))
            self.host = host
            self.is_connected_explicit = True
        except socket.error as err:
            if self.explicit_message_socket:
                await self.explicit_message_socket.close()
                self.is_connected_explicit = False
            raise err

    async def close_explicit(self):
        """
        Close the explicit Ethernet/IP connection
        :return:
        """
        self.is_connected_explicit = False
        self.has_session_handle = False
        self.host = None
        if self.explicit_message_socket:
            await self.explicit_message_socket.close()

    async def register_session(self, command_data=b'\x01\x00\x00\x00'):
        """
        Used by an originator to establish a session. It is required before sending CIP messages
        :param command_data:
        :return:
        """
        eip_message = EIPMessage(b'\x65\x00', command_data)
        response = await self.send_command(eip_message, self.host)
        self.has_session_handle = True
        self.session_handle_id = response.session_handle_id
        return response


class EIPConnectedCommandMixin(EIPDispatcher):
    def __init__(self):
        super().__init__()
        self.explicit_message_socket = None
        self.session_handle_id = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.is_connected_explicit = False
        self.has_session_handle = False
        self.BUFFER_SIZE = 4096
        self.host = None

    def __del__(self):
        self.close_explicit()

    def get_response(self, eip_message: EIPMessage):
        if eip_message.context_integer() in self.eip_responses.keys():
            result = self.eip_responses.pop(eip_message.context_integer())
            return result
        else:
            self.explicit_message_socket.send(eip_message.bytes())
            received_data = self.explicit_message_socket.recv(self.BUFFER_SIZE)
            length = struct.unpack('<H',received_data[2:4])[0] + 24
            received_eip_message = EIPMessage()
            while len(received_data) > length:
                received_eip_message.from_bytes(received_data[0:length])
                received_data = received_data[length:]
                length = struct.unpack('<H', received_data[2:4])[0] + 24
                self.eip_responses[received_eip_message.context_integer()] = received_eip_message
            received_eip_message.from_bytes(received_data[0:length])
            if len(received_data) > 0:
                received_eip_message.from_bytes(received_data)
            self.eip_responses[received_eip_message.context_integer()] = received_eip_message
            return self.get_response(eip_message)


    def send_command(self, eip_command: EIPMessage, host) -> EIPMessage:
        """
        Used to send and receive Ethernet/IP messages
        :param host:
        :param eip_command:
        :return:
        """
        # received_eip_message = EIPMessage()
        if self.is_connected_explicit:
            if self.message_number == 18446744073709551616:
                self.message_number = 0
            eip_command.set_context(self.message_number)
            self.message_number = self.message_number + 1
            return self.get_response(eip_command)

    def connect_explicit(self, host, connection_timeout: float = None):
        """
        Create and explicit Ethernet/IP connection
        :param connection_timeout:
        :param host:
        """
        try:
            self.explicit_message_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            old_timeout = self.explicit_message_socket.gettimeout()
            if connection_timeout is not None:
                self.explicit_message_socket.settimeout(connection_timeout)
            self.explicit_message_socket.connect((host, self.explicit_message_port))
            self.explicit_message_socket.settimeout(old_timeout)
            self.host = host
            self.is_connected_explicit = True
        except socket.error as err:
            if self.explicit_message_socket:
                self.explicit_message_socket.close()
                self.is_connected_explicit = False
            raise err

    def close_explicit(self):
        """
        Close the explicit Ethernet/IP connection
        :return:
        """
        self.is_connected_explicit = False
        self.has_session_handle = False
        self.host = None
        if self.explicit_message_socket:
            self.explicit_message_socket.close()

    def register_session(self, command_data=b'\x01\x00\x00\x00'):
        """
        Used by an originator to establish a session. It is required before sending CIP messages
        :param command_data:
        :return:
        """
        eip_message = EIPMessage(b'\x65\x00', command_data)
        response = self.send_command(eip_message, self.host)
        self.has_session_handle = True
        self.session_handle_id = response.session_handle_id
        return response


class EIPUnconnectedCommandMixin(EIPDispatcher):
    """
    ToDo add broadcast_command to return a List of EIPMessages (List Service to a full domain)
    """
    def __init__(self):
        super().__init__()

    def send_command(self, eip_command: EIPMessage, host: str) -> List[EIPMessage]:
        received_eip_message = EIPMessage()
        response_list = []
        try:
            udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            udp_socket.settimeout(5)
            print(self.explicit_message_port)
            udp_socket.bind(('0.0.0.0', self.explicit_message_port))
            udp_socket.sendto(eip_command.bytes(), (host, self.explicit_message_port))
            received_data = udp_socket.recvfrom(1024)[0]
            # while True:
            #     try:
            #         received_data = udp_socket.recvfrom(1024)[0]
            #         # response_list.append(received_eip_message.from_bytes(received_data))
            #     except socket.error as e:
            #         break
            received_eip_message.from_bytes(received_data)
        except socket.error as e:
            print('Failed to create socket')
            print(e)

        return received_eip_message


class AsyncEIPConnectedCIPDispatcher(AsyncEIPConnectedCommandMixin, AsyncCIPDispatcher):
    """
    EIP is an encapsulation protocol for CIP (common industrial protocol) messages
    """

    null_address_item = b'\x00\x00\x00\x00'
    cip_handle = b'\x00\x00\x00\x00'

    def __init__(self):
        super().__init__()

    async def execute_cip_command(self, request: CIPRequest) -> CIPReply:
        """
        Implements the abstract method in order to become a concrete CIPDispatcher class
        :param request:
        :return:
        """
        data_address_item = DataAndAddressItem(DataAndAddressItem.UNCONNECTED_MESSAGE, request.bytes)
        packets = [data_address_item]
        # ToDo add interface handle to track responses?
        common_packet_format = CommonPacketFormat(packets)
        command_specific_data = CommandSpecificData(encapsulated_packet=common_packet_format.bytes())
        response = await self.send_rr_data(command_specific_data.bytes())
        response = response.packets[1].bytes()
        reply_data_and_address_item = DataAndAddressItem('', b'')
        # ToDo is this removing error data?
        reply_data_and_address_item.from_bytes(response)
        cip_reply = CIPReply(reply_data_and_address_item.data)
        if cip_reply.general_status != b'\x00':
            raise CIPException(cip_reply.general_status, cip_reply.extended_status)
        return cip_reply


    async def send_rr_data(self, command_specific_data: bytes) -> CommonPacketFormat:
        """
        Ethernet/IP command to send an encapsulated request and reply packet between originator and target
        :param command_specific_data:
        :return:
        """
        eip_message = EIPMessage(b'\x6f\x00', command_specific_data, self.session_handle_id)
        reply = await self.send_command(eip_message, self.host)
        reply_command_specific_data = CommandSpecificData()
        reply_command_specific_data.from_bytes(reply.command_data)
        reply_packet = CommonPacketFormat([])
        reply_packet.from_bytes(reply_command_specific_data.encapsulated_packet)
        return reply_packet


class EIPConnectedCIPDispatcher(EIPConnectedCommandMixin, CIPDispatcher):
    """
    EIP is an encapsulation protocol for CIP (common industrial protocol) messages
    """

    null_address_item = b'\x00\x00\x00\x00'
    cip_handle = b'\x00\x00\x00\x00'

    def __init__(self):
        super().__init__()

    def execute_cip_command(self, request: CIPRequest) -> CIPReply:
        """
        Implements the abstract method in order to become a concrete CIPDispatcher class
        :param request:
        :return:
        """
        data_address_item = DataAndAddressItem(DataAndAddressItem.UNCONNECTED_MESSAGE, request.bytes)
        packets = [data_address_item]
        # ToDo add interface handle to track responses?
        common_packet_format = CommonPacketFormat(packets)
        command_specific_data = CommandSpecificData(encapsulated_packet=common_packet_format.bytes())
        response = self.send_rr_data(command_specific_data.bytes()).packets[1].bytes()
        reply_data_and_address_item = DataAndAddressItem('', b'')
        # ToDo is this removing error data?
        reply_data_and_address_item.from_bytes(response)
        cip_reply = CIPReply(reply_data_and_address_item.data)
        if cip_reply.general_status != b'\x00':
            raise CIPException(cip_reply.general_status, cip_reply.extended_status)
        return cip_reply

    # @staticmethod
    # def command_specific_data_from_eip_message_bytes(eip_message_bytes: bytes):
    #     """
    #     Extracts command specific data from Ethernet/IP reply bytes
    #     :param eip_message_bytes:
    #     :return:
    #     """
    #     eip_message = EIPMessage()
    #     eip_message.from_bytes(eip_message_bytes)
    #     return EIPConnectedCIPDispatcher.command_specific_data_from_eip_message(eip_message)
    #
    # @staticmethod
    # def command_specific_data_from_eip_message(eip_message: EIPMessage) -> CommandSpecificData:
    #     """
    #     Extracts the command specific data from an Ethernet/IP message
    #     :param eip_message:
    #     :return:
    #     """
    #     command_specific_data = CommandSpecificData()
    #     command_specific_data.from_bytes(eip_message.bytes())
    #     return command_specific_data

    def send_rr_data(self, command_specific_data: bytes) -> CommonPacketFormat:
        """
        Ethernet/IP command to send an encapsulated request and reply packet between originator and target
        :param command_specific_data:
        :return:
        """
        eip_message = EIPMessage(b'\x6f\x00', command_specific_data, self.session_handle_id)
        reply = self.send_command(eip_message, self.host)
        reply_command_specific_data = CommandSpecificData()
        reply_command_specific_data.from_bytes(reply.command_data)
        reply_packet = CommonPacketFormat([])
        reply_packet.from_bytes(reply_command_specific_data.encapsulated_packet)
        return reply_packet

    # def get_attribute_all_service(self, tag_request_path):
    #     get_attribute_all_request = CIPRequest(CIPService.GET_ATTRIBUTE_ALL, tag_request_path)
    #     return self.execute_cip_command(get_attribute_all_request)
    #
    # def get_attribute_single_service(self, tag_request_path):
    #     get_attribute_single_request = CIPRequest(CIPService.GET_ATTRIBUTE_SINGLE, tag_request_path)
    #     return self.execute_cip_command(get_attribute_single_request)



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from aphyt.eip import *
from decimal import *
from aphyt.cip.cip_datatypes import *
from aphyt.cip.cip import *


class V4Series:
    def __init__(self):
        super().__init__()
        self.connected_cip_dispatcher = EIPConnectedCIPDispatcher()
        self.last_read_length = 0
        self.last_read_string = ''

    def connect_explicit(self, host):
        # ToDo Consider moving all these convenience methods to a Mixin
        self.connected_cip_dispatcher.connect_explicit(host)

    def close_explicit(self):
        self.connected_cip_dispatcher.close_explicit()

    def register_session(self):
        self.connected_cip_dispatcher.register_session()

    def execute_command(self, command) -> bytes:
        service_code = b'\x45'
        request_path = address_request_path_segment(class_id=b'\x68\x00', instance_id=b'\x01', attribute_id=b'\x01')
        data = struct.pack('<I', len(command))
        data = data + bytes(command, 'utf-8')
        request = CIPRequest(service_code, request_path, request_data=data)
        reply = self.connected_cip_dispatcher.execute_cip_command(request)
        return reply.reply_data

    def read_execute(self, delimiter=' '):
        command = '<' + delimiter + '>'
        reply = self.execute_command(command)
        length = struct.unpack('<I', reply[0:4])[0]
        self.last_read_length = length
        self.last_read_string = str(reply[4:], 'utf-8')
        return self.last_read_string





